#coding=utf8

################################################################################
###                                                                          ###
### Created by Mahdi Manoochehrtayebi, 2020-2024                             ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
###                                                                          ###
### And Martin Genet, 2018-2024                                              ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin
import gmsh
import meshio
# import mshr

import dolfin_mech as dmech

################################################################################

def setPeriodic(coord, xmin, ymin, zmin, xmax, ymax, zmax, e=1e-6):
    # From https://gitlab.onelab.info/gmsh/gmsh/-/issues/744
    
    smin = gmsh.model.getEntitiesInBoundingBox(
        xmin - e,
        ymin - e,
        zmin - e,
        (xmin + e) if (coord == 0) else (xmax + e),
        (ymin + e) if (coord == 1) else (ymax + e),
        (zmin + e) if (coord == 2) else (zmax + e),
        2)
    dx = (xmax - xmin) if (coord == 0) else 0
    dy = (ymax - ymin) if (coord == 1) else 0
    dz = (zmax - zmin) if (coord == 2) else 0

    for i in smin:
        bb = gmsh.model.getBoundingBox(i[0], i[1])
        bbe = [bb[0] - e + dx, bb[1] - e + dy, bb[2] - e + dz,
               bb[3] + e + dx, bb[4] + e + dy, bb[5] + e + dz]
        smax = gmsh.model.getEntitiesInBoundingBox(
            bbe[0], bbe[1], bbe[2],
            bbe[3], bbe[4], bbe[5])
        for j in smax:
            bb2 = list(gmsh.model.getBoundingBox(j[0], j[1]))
            bb2[0] -= dx; bb2[1] -= dy; bb2[2] -= dz;
            bb2[3] -= dx; bb2[4] -= dy; bb2[5] -= dz;
            if ((abs(bb2[0] - bb[0]) < e) and (abs(bb2[1] - bb[1]) < e) and
                (abs(bb2[2] - bb[2]) < e) and (abs(bb2[3] - bb[3]) < e) and
                (abs(bb2[4] - bb[4]) < e) and (abs(bb2[5] - bb[5]) < e)):
                gmsh.model.mesh.setPeriodic(
                    2,
                    [j[1]], [i[1]],
                    [1, 0, 0, dx,\
                     0, 1, 0, dy,\
                     0, 0, 1, dz,\
                     0, 0, 0, 1 ])

################################################################################

def run_HollowBox_Mesh(
        params : dict = {}):

    dim       = params.get("dim"); assert (dim in (2,3))
    xmin      = params.get("xmin", 0.)
    ymin      = params.get("ymin", 0.)
    zmin      = params.get("zmin", 0.)
    xmax      = params.get("xmax", 1.)
    ymax      = params.get("ymax", 1.)
    zmax      = params.get("zmax", 1.)
    xshift    = params.get("xshift", 0.)
    yshift    = params.get("yshift", 0.)
    zshift    = params.get("zshift", 0.)
    r0        = params.get("r0", 0.2)
    l         = params.get("l", 0.1)

    mesh_filebasename = params.get("mesh_filebasename", "mesh")

    ################################################################### Mesh ###

    gmsh.initialize()

    if (dim==2):
        box_tag   = 1
        hole1_tag = 2
        hole2_tag = 3
        hole3_tag = 4
        hole4_tag = 5
        rve_tag   = 6

        gmsh.model.occ.addRectangle(x=xmin+xshift, y=ymin+yshift, z=0., dx=xmax-xmin, dy=ymax-ymin, tag=box_tag)
        gmsh.model.occ.addDisk(xc=xmin, yc=ymin, zc=0., rx=r0, ry=r0, tag=hole1_tag)
        gmsh.model.occ.addDisk(xc=xmax, yc=ymin, zc=0., rx=r0, ry=r0, tag=hole2_tag)
        gmsh.model.occ.addDisk(xc=xmax, yc=ymax, zc=0., rx=r0, ry=r0, tag=hole3_tag)
        gmsh.model.occ.addDisk(xc=xmin, yc=ymax, zc=0., rx=r0, ry=r0, tag=hole4_tag)
        gmsh.model.occ.cut(objectDimTags=[(2, box_tag)], toolDimTags=[(2, hole1_tag), (2, hole2_tag), (2, hole3_tag), (2, hole4_tag)], tag=rve_tag)
        gmsh.model.occ.synchronize()
        gmsh.model.addPhysicalGroup(dim=2, tags=[rve_tag])
        dmech.setPeriodic(coord=0, xmin=xmin, ymin=ymin, zmin=zmin, xmax=xmax, ymax=ymax, zmax=zmax)
        dmech.setPeriodic(coord=1, xmin=xmin, ymin=ymin, zmin=zmin, xmax=xmax, ymax=ymax, zmax=zmax)
        # gmsh.model.mesh.setPeriodic(dim=1, tags=[2], tagsMaster=[1], affineTransform=[1, 0, 0, xmax-xmin,\
        #                                                                               0, 1, 0, 0        ,\
        #                                                                               0, 0, 1, 0        ,\
        #                                                                               0, 0, 0, 1        ])
        # gmsh.model.mesh.setPeriodic(dim=1, tags=[4], tagsMaster=[3], affineTransform=[1, 0, 0, 0        ,\
        #                                                                               0, 1, 0, ymax-ymin,\
        #                                                                               0, 0, 1, 0        ,\
        #                                                                               0, 0, 0, 1        ])
        gmsh.model.mesh.setSize(dimTags=gmsh.model.getEntities(0), size=l)
        gmsh.model.mesh.generate(dim=2)
    if (dim==3):
        box_tag   = 1
        hole1_tag = 2
        hole2_tag = 3
        hole3_tag = 4
        hole4_tag = 5
        hole5_tag = 6
        hole6_tag = 7
        hole7_tag = 8
        hole8_tag = 9
        rve_tag   = 10

        gmsh.model.occ.addBox(x=xmin+xshift, y=ymin+yshift, z=zmin+zshift, dx=xmax-xmin, dy=ymax-ymin, dz=zmax-zmin, tag=box_tag)
        gmsh.model.occ.addSphere(xc=xmin, yc=ymin, zc=zmin, radius=r0, tag=hole1_tag)
        gmsh.model.occ.addSphere(xc=xmax, yc=ymin, zc=zmin, radius=r0, tag=hole2_tag)
        gmsh.model.occ.addSphere(xc=xmax, yc=ymax, zc=zmin, radius=r0, tag=hole3_tag)
        gmsh.model.occ.addSphere(xc=xmin, yc=ymax, zc=zmin, radius=r0, tag=hole4_tag)
        gmsh.model.occ.addSphere(xc=xmin, yc=ymin, zc=zmax, radius=r0, tag=hole5_tag)
        gmsh.model.occ.addSphere(xc=xmax, yc=ymin, zc=zmax, radius=r0, tag=hole6_tag)
        gmsh.model.occ.addSphere(xc=xmax, yc=ymax, zc=zmax, radius=r0, tag=hole7_tag)
        gmsh.model.occ.addSphere(xc=xmin, yc=ymax, zc=zmax, radius=r0, tag=hole8_tag)
        gmsh.model.occ.cut(objectDimTags=[(3, box_tag)], toolDimTags=[(3, hole1_tag), (3, hole2_tag), (3, hole3_tag), (3, hole4_tag), (3, hole5_tag), (3, hole6_tag), (3, hole7_tag), (3, hole8_tag)], tag=rve_tag)
        gmsh.model.occ.synchronize()
        gmsh.model.addPhysicalGroup(dim=3, tags=[rve_tag])
        setPeriodic(coord=0, xmin=xmin, ymin=ymin, zmin=zmin, xmax=xmax, ymax=ymax, zmax=zmax)
        setPeriodic(coord=1, xmin=xmin, ymin=ymin, zmin=zmin, xmax=xmax, ymax=ymax, zmax=zmax)
        setPeriodic(coord=2, xmin=xmin, ymin=ymin, zmin=zmin, xmax=xmax, ymax=ymax, zmax=zmax)
        # gmsh.model.mesh.setPeriodic(dim=2, tags=[2], tagsMaster=[1], affineTransform=[1, 0, 0, xmax-xmin,\
        #                                                                               0, 1, 0, 0        ,\
        #                                                                               0, 0, 1, 0        ,\
        #                                                                               0, 0, 0, 1        ])
        # gmsh.model.mesh.setPeriodic(dim=2, tags=[4], tagsMaster=[3], affineTransform=[1, 0, 0, 0        ,\
        #                                                                               0, 1, 0, ymax-ymin,\
        #                                                                               0, 0, 1, 0        ,\
        #                                                                               0, 0, 0, 1        ])
        # gmsh.model.mesh.setPeriodic(dim=2, tags=[6], tagsMaster=[5], affineTransform=[1, 0, 0, 0        ,\
        #                                                                               0, 1, 0, 0        ,\
        #                                                                               0, 0, 1, zmax-zmin,\
        #                                                                               0, 0, 0, 1        ])
        gmsh.model.mesh.setSize(dimTags=gmsh.model.getEntities(0), size=l)
        gmsh.model.mesh.generate(dim=3)

    gmsh.write(mesh_filebasename+"-mesh.vtk")
    gmsh.finalize()

    mesh = meshio.read(mesh_filebasename+"-mesh.vtk")
    if (dim==2): mesh.points = mesh.points[:, :2]
    meshio.write(mesh_filebasename+"-mesh.xdmf", mesh)

    mesh = dolfin.Mesh()
    dolfin.XDMFFile(mesh_filebasename+"-mesh.xdmf").read(mesh)
    
    # if   (dim==2):
    #     geometry = mshr.Rectangle(dolfin.Point(xmin, ymin), dolfin.Point(xmax, xmax))\
    #              - mshr.Circle(dolfin.Point(x0, y0), r0)
    # elif (dim==3):
    #     geometry = mshr.Box(dolfin.Point(xmin, ymin, zmin), dolfin.Point(xmax, ymax, zmax))\
    #              - mshr.Sphere(dolfin.Point(x0, y0, z0), r0)
    # mesh = mshr.generate_mesh(geometry, 10)

    return mesh



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin

################################################################################

def run_RivlinCube_Mesh(
        dim    : int  = 3 ,
        params : dict = {}):

    X0 = params.get("X0", 0.)
    X1 = params.get("X1", 1.)
    Y0 = params.get("Y0", 0.)
    Y1 = params.get("Y1", 1.)
    if (dim==3): Z0 = params.get("Z0", 0.)
    if (dim==3): Z1 = params.get("Z1", 1.)
    l = params.get("l", 1.)

    mesh_filebasename = params.get("mesh_filebasename", "mesh")

    LX = X1-X0
    LY = Y1-Y0
    if (dim==3): LZ = Z1-Z0

    NX = int(LX/l)
    NY = int(LY/l)
    if (dim==3): NZ = int(LZ/l)

    if (dim==2):
        mesh = dolfin.RectangleMesh(
            dolfin.Point(X0, Y0, 0.), dolfin.Point(Y1, Y1, 0.),
            NX, NY,
            "crossed")
    elif (dim==3):
        mesh = dolfin.BoxMesh(
            dolfin.Point(X0, Y0, Z0), dolfin.Point(X1, Y1, Z1),
            NX, NY, NZ)

    xdmf_file_mesh = dolfin.XDMFFile(mesh_filebasename+".xdmf")
    xdmf_file_mesh.write(mesh)
    xdmf_file_mesh.close()

    dolfin.File(mesh_filebasename+".xml") << mesh

    ################################################## Subdomains & Measures ###

    xmin_sd = dolfin.CompiledSubDomain("near(x[0], x0) && on_boundary", x0=X0)
    xmax_sd = dolfin.CompiledSubDomain("near(x[0], x0) && on_boundary", x0=X1)
    ymin_sd = dolfin.CompiledSubDomain("near(x[1], x0) && on_boundary", x0=Y0)
    ymax_sd = dolfin.CompiledSubDomain("near(x[1], x0) && on_boundary", x0=Y1)
    if (dim==3): zmin_sd = dolfin.CompiledSubDomain("near(x[2], x0) && on_boundary", x0=Z0)
    if (dim==3): zmax_sd = dolfin.CompiledSubDomain("near(x[2], x0) && on_boundary", x0=Z1)

    xmin_id = 1
    xmax_id = 2
    ymin_id = 3
    ymax_id = 4
    if (dim==3): zmin_id = 5
    if (dim==3): zmax_id = 6

    boundaries_mf = dolfin.MeshFunction("size_t", mesh, mesh.topology().dim()-1) # MG20180418: size_t looks like unisgned int, but more robust wrt architecture and os
    boundaries_mf.set_all(0)
    xmin_sd.mark(boundaries_mf, xmin_id)
    xmax_sd.mark(boundaries_mf, xmax_id)
    ymin_sd.mark(boundaries_mf, ymin_id)
    ymax_sd.mark(boundaries_mf, ymax_id)
    if (dim==3): zmin_sd.mark(boundaries_mf, zmin_id)
    if (dim==3): zmax_sd.mark(boundaries_mf, zmax_id)

    # xdmf_file_boundaries = dolfin.XDMFFile("boundaries.xdmf")
    # xdmf_file_boundaries.write(boundaries_mf)
    # xdmf_file_boundaries.close()

    if (dim==2):
        return mesh, boundaries_mf, xmin_id, xmax_id, ymin_id, ymax_id
    elif (dim==3):
        return mesh, boundaries_mf, xmin_id, xmax_id, ymin_id, ymax_id, zmin_id, zmax_id



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin

import dolfin_mech as dmech
from .Problem_Hyperelasticity_Poro import PoroHyperelasticityProblem

################################################################################

class InversePoroHyperelasticityProblem(PoroHyperelasticityProblem):



    def __init__(self, *args, **kwargs):

        PoroHyperelasticityProblem.__init__(self, *args, **kwargs)



    def get_displacement_name(self):

        return "u"



    def get_porosity_name(self):

        return "phis0"



    def set_kinematics(self):

        self.kinematics = dmech.InverseKinematics(
            u=self.get_displacement_subsol().subfunc,
            u_old=self.get_displacement_subsol().func_old)

        self.add_foi(expr=self.kinematics.F, fs=self.mfoi_fs, name="F")
        self.add_foi(expr=self.kinematics.J, fs=self.sfoi_fs, name="J")
        self.add_foi(expr=self.kinematics.C, fs=self.mfoi_fs, name="C")
        self.add_foi(expr=self.kinematics.E, fs=self.mfoi_fs, name="E")



    def init_known_porosity(self,
            porosity_init_val,
            porosity_init_fun):

        if   (porosity_init_val   is not None):
            self.phis = dolfin.Constant(porosity_init_val)
        elif (porosity_init_fun is not None):
            self.phis = porosity_init_fun
        self.add_foi(
            expr=self.phis,
            fs=self.get_porosity_function_space().collapse(),
            name="phis")
        self.add_foi(
            expr=1 - self.phis,
            fs=self.get_porosity_function_space().collapse(),
            name="phif")
        self.add_foi(
            expr=1/self.kinematics.J - self.get_porosity_subsol().subfunc,
            fs=self.get_porosity_function_space().collapse(),
            name="phif0")
        self.add_foi(
            expr=self.kinematics.J * self.get_porosity_subsol().subfunc,
            fs=self.get_porosity_function_space().collapse(),
            name="Phis0")
        self.add_foi(
            expr=1 - self.kinematics.J * self.get_porosity_subsol().subfunc,
            fs=self.get_porosity_function_space().collapse(),
            name="Phif0")



    def add_Wskel_operator(self,
            material_parameters,
            material_scaling,
            subdomain_id=None):

        operator = dmech.InverseWskelPoroOperator(
            kinematics=self.kinematics,
            u_test=self.get_displacement_subsol().dsubtest,
            phis0=self.get_porosity_subsol().subfunc,
            material_parameters=material_parameters,
            material_scaling=material_scaling,
            measure=self.get_subdomain_measure(subdomain_id))
        return self.add_operator(operator)



    def add_Wbulk_operator(self,
            material_parameters,
            material_scaling,
            subdomain_id=None):

        operator = dmech.InverseWbulkPoroOperator(
            kinematics=self.kinematics,
            u_test=self.get_displacement_subsol().dsubtest,
            phis=self.phis,
            phis0=self.get_porosity_subsol().subfunc,
            phis0_test=self.get_porosity_subsol().dsubtest,
            material_parameters=material_parameters,
            material_scaling=material_scaling,
            measure=self.get_subdomain_measure(subdomain_id))
        return self.add_operator(operator)



    def add_Wpore_operator(self,
            material_parameters,
            material_scaling,
            subdomain_id=None):

        operator = dmech.InverseWporePoroOperator(
            kinematics=self.kinematics,
            phis=self.phis,
            phis0=self.get_porosity_subsol().subfunc,
            phis0_test=self.get_porosity_subsol().dsubtest,
            material_parameters=material_parameters,
            material_scaling=material_scaling,
            measure=self.get_subdomain_measure(subdomain_id))
        return self.add_operator(operator)



    def add_global_porosity_qois(self):

        self.add_qoi(
            name=self.get_porosity_name(),
            expr=self.get_porosity_subsol().subfunc * self.dV)

        self.add_qoi(
            name="phif0",
            expr=(1/self.kinematics.J - self.get_porosity_subsol().subfunc) * self.dV)

        self.add_qoi(
            name="Phis0",
            expr=(self.kinematics.J * self.get_porosity_subsol().subfunc) * self.dV)

        self.add_qoi(
            name="Phif0",
            expr=(1 - self.kinematics.J * self.get_porosity_subsol().subfunc) * self.dV)



#coding=utf8

##############################################################################
###                                                                        ###
### Created by Mahdi Manoocherhtayebi 2020-2024                            ###
###                                                                        ###  
### Inspired by https://comet-fenics.readthedocs.io/en/latest/demo/        ###
###                         periodic_homog_elas/periodic_homog_elas.html   ###
###                                                                        ###
### École Polytechnique, Palaiseau, France                                 ###
###                                                                        ###
##############################################################################

import dolfin
import numpy

import dolfin_mech as dmech

#############################################################################

class HomogenizationProblem():



    def __init__(self,
            dim,
            mesh,
            mat_params,
            vol,
            bbox,
            vertices=None):

        self.mesh = mesh
        self.E_s = mat_params["E"]
        self.nu_s = mat_params["nu"]
        self.dim = dim
        self.vertices = vertices
        self.vol = vol

        self.bbox = bbox
        # print("self.vol:", self.vol)
    
        self.material_parameters = [(self.E_s, self.nu_s)]



    def eps(self, v):

        return dolfin.sym(dolfin.grad(v))



    def sigma(self, v, i, Eps):

        E, nu = self.material_parameters[i]
        lmbda = dolfin.Constant(E*nu/(1+nu)/(1-2*nu))
        mu = dolfin.Constant(E/2/(1+nu))
        return lmbda*dolfin.tr(self.eps(v)+Eps)*dolfin.Identity(self.dim) + 2*mu*(self.eps(v)+Eps)



    def Voigt2strain(self, s):

        if (self.dim==2):
            strain_tensor = numpy.array([[s[0]   , s[2]/2.],
                                         [s[2]/2., s[1]   ]])
        if (self.dim==3):
            strain_tensor = numpy.array([[s[0]   , s[5]/2.,  s[4]/2],
                                         [s[5]/2., s[1]   ,  s[3]/2],
                                         [s[4]/2 , s[3]/2 ,  s[2]  ]])
        return strain_tensor



    def get_macro_strain(self, i):
        """returns the macroscopic strain for the 3 elementary load cases"""

        if (self.dim==2):
            Eps_Voigt = numpy.zeros(3)
        if (self.dim==3):
            Eps_Voigt = numpy.zeros(6)
        Eps_Voigt[i] = 1
        return self.Voigt2strain(Eps_Voigt)



    def stress2Voigt(self, s):

        if (self.dim==2):
            stress_vector = dolfin.as_vector([s[0,0], s[1,1], s[0,1]])
        if (self.dim==3):
            stress_vector = dolfin.as_vector([s[0,0], s[1,1], s[2,2], s[1,2], s[0,2], s[0,1]])
        return stress_vector



    def get_lambda_and_mu(self):

        Ve = dolfin.VectorElement("CG", self.mesh.ufl_cell(), 2)
        Re = dolfin.VectorElement("R", self.mesh.ufl_cell(), 0)
        W = dolfin.FunctionSpace(self.mesh, dolfin.MixedElement([Ve, Re]), constrained_domain=dmech.PeriodicSubDomain(self.dim, self.bbox, self.vertices))
        V = dolfin.FunctionSpace(self.mesh, Ve)

        v_, lamb_ = dolfin.TestFunctions(W)
        dv, dlamb = dolfin.TrialFunctions(W)
        w = dolfin.Function(W)
        dx = dolfin.Measure('dx')(domain=self.mesh)

        if (self.dim==2): Eps = dolfin.Constant(((0, 0), (0, 0)))
        if (self.dim==3): Eps = dolfin.Constant(((0, 0, 0), (0, 0, 0), (0, 0, 0)))

        F = sum([dolfin.inner(self.sigma(dv, 0, Eps), self.eps(v_))*dx])

        a, L = dolfin.lhs(F), dolfin.rhs(F)
        a += dolfin.dot(lamb_,dv)*dx + dolfin.dot(dlamb,v_)*dx

        if (self.dim==2): Chom = numpy.zeros((3, 3))
        if (self.dim==3): Chom = numpy.zeros((6, 6))

        if (self.dim==2): Enum=enumerate(["Exx", "Eyy", "Exy"])
        if (self.dim==3): Enum=enumerate(["Exx", "Eyy", "Ezz", "Eyz", "Exz", "Exy"])

        for (j, case) in Enum:
            # print("Solving {} case...".format(case))
            macro_strain = self.get_macro_strain(j)
            Eps.assign(dolfin.Constant(macro_strain))
            if (self.dim ==3):
                dolfin.solve(a == L, w, [], solver_parameters={"linear_solver": "cg"})
            else:
                dolfin.solve(a == L, w, [], solver_parameters={"linear_solver": "lu"})
            (v, lamb) = dolfin.split(w)
            # xdmf_file_per.write(w, float(j))

            if (self.dim==2): Sigma_len = 3
            if (self.dim==3): Sigma_len = 6

            Sigma = numpy.zeros((Sigma_len,))
            
            # self.vol = dolfin.assemble(dolfin.Constant(1) * dx)
            for k in range(Sigma_len):
                Sigma[k] = dolfin.assemble(self.stress2Voigt(self.sigma(v, 0, Eps))[k]*dx)/self.vol
            Chom[j, :] = Sigma
        
        # print(Chom)
        
        lmbda_hom = Chom[0, 1]
        if (self.dim==2): mu_hom = Chom[2, 2]
        if (self.dim==3): mu_hom = Chom[4, 4]

        # print(Chom[0, 0], lmbda_hom + 2*mu_hom)
        # print('Chom:' +str(Chom))

        # print("lmbda_hom: " +str(lmbda_hom))
        # print("mu_hom: " +str(mu_hom))

        E_hom = mu_hom*(3*lmbda_hom + 2*mu_hom)/(lmbda_hom + mu_hom)
        nu_hom = lmbda_hom/(lmbda_hom + mu_hom)/2

        # print("E_hom: " +str(E_hom))
        # print("nu_hom: " +str(nu_hom))
        # print(Chom[0, 0], lmbda_hom + 2*mu_hom)
        # print("Isotropy = " +str ((lmbda_hom + 2*mu_hom - abs(lmbda_hom + 2*mu_hom - Chom[0, 0]))/(lmbda_hom + 2*mu_hom) * 100) +"%")

        self.E_hom = E_hom
        self.nu_hom = nu_hom 
        return lmbda_hom, mu_hom



    def get_kappa(self):

        p_f = 1

        coord = self.mesh.coordinates()
        xmax = max(coord[:,0]); xmin = min(coord[:,0])
        ymax = max(coord[:,1]); ymin = min(coord[:,1])
        if (self.dim==3): zmax = max(coord[:,2]); zmin = min(coord[:,2])
        if (self.dim==2):    
            vol = (xmax - xmin)*(ymax - ymin)

        if (self.dim==3):    
            vol = (xmax - xmin)*(ymax - ymin)*(zmax - zmin)

        vertices = numpy.array([[xmin, ymin],
                                [xmax, ymin],
                                [xmax, ymax],
                                [xmin, ymax]])
            
        tol = 1E-8
        vv = vertices
        a1 = vv[1,:]-vv[0,:] # first vector generating periodicity
        a2 = vv[3,:]-vv[0,:] # second vector generating periodicity
        # check if UC vertices form indeed a parallelogram
        assert numpy.linalg.norm(vv[2, :]-vv[3, :] - a1) <= tol
        assert numpy.linalg.norm(vv[2, :]-vv[1, :] - a2) <= tol

        ################################################## Subdomains & Measures ###

        class BoundaryX0(dolfin.SubDomain):
            def inside(self,x,on_boundary):
                # return on_boundary and dolfin.near(x[0],xmin,tol)
                return on_boundary and dolfin.near(x[0],vv[0,0] + x[1]*a2[0]/vv[3,1],tol)

        class BoundaryY0(dolfin.SubDomain):
            def inside(self,x,on_boundary):
                return on_boundary and dolfin.near(x[1],ymin,tol)

        class BoundaryX1(dolfin.SubDomain):
            def inside(self,x,on_boundary):
                # return on_boundary and dolfin.near(x[0],xmax,tol)
                return on_boundary and dolfin.near(x[0],vv[1,0] + x[1]*a2[0]/vv[3,1],tol)

        class BoundaryY1(dolfin.SubDomain):
            def inside(self,x,on_boundary):
                return on_boundary and dolfin.near(x[1],ymax,tol)            

        if (self.dim ==3):
            class BoundaryZ0(dolfin.SubDomain):
                def inside(self,x,on_boundary):
                    # return on_boundary and dolfin.near(x[0],xmax,tol)
                    return on_boundary and dolfin.near(x[2],zmin,tol)

            class BoundaryZ1(dolfin.SubDomain):
                def inside(self,x,on_boundary):
                    return on_boundary and dolfin.near(x[2],zmax,tol)

        boundaries_mf = dolfin.MeshFunction("size_t", self.mesh, self.mesh.topology().dim() - 1)
        boundaries_mf.set_all(0)

        bX0 = BoundaryX0()
        bY0 = BoundaryY0()
        bX1 = BoundaryX1()
        bY1 = BoundaryY1()
        if (self.dim==3):
            bZ0 = BoundaryZ0()
            bZ1 = BoundaryZ1()

        bX0.mark(boundaries_mf, 1)
        bY0.mark(boundaries_mf, 2)
        bX1.mark(boundaries_mf, 3)
        bY1.mark(boundaries_mf, 4)
        if (self.dim==3):
            bZ0.mark(boundaries_mf, 5)
            bZ1.mark(boundaries_mf, 6)

        dV = dolfin.Measure(
            "dx",
            domain=self.mesh)
        dS = dolfin.Measure(
            "exterior_facet",
            domain=self.mesh,
            subdomain_data=boundaries_mf)

        ############################################################# Functions #######

        def eps(v_bar):
            return dolfin.sym(dolfin.grad(v_bar))

        def sigma(v_bar, i, Eps):
            E, nu = self.material_parameters[i]
            lmbda = E*nu/(1+nu)/(1-2*nu)
            mu = E/2/(1+nu)
            return lmbda*dolfin.tr(eps(v_bar) + Eps)*dolfin.Identity(self.dim) + 2*mu*(eps(v_bar)+Eps)

        Ve = dolfin.VectorElement("CG", self.mesh.ufl_cell(), 2)
        Re = dolfin.VectorElement("R", self.mesh.ufl_cell(), 0)
        W = dolfin.FunctionSpace(self.mesh, dolfin.MixedElement([Ve, Re]), constrained_domain=dmech.PeriodicSubDomain(self.dim, self.bbox, vertices))
        V = dolfin.FunctionSpace(self.mesh, Ve)

        v_, lamb_ = dolfin.TestFunctions(W)
        dv, dlamb = dolfin.TrialFunctions(W)
        w = dolfin.Function(W)

        (v, lamb) = dolfin.split(w)       

        mesh_V0 = dolfin.assemble(dolfin.Constant(1) * dV)

        ##################################################################################
        
        macro_strain = numpy.zeros((self.dim, self.dim))

        if (self.dim==2): Eps = dolfin.Constant(((0, 0), (0, 0)))
        if (self.dim==3): Eps = dolfin.Constant(((0, 0, 0), (0, 0, 0), (0, 0, 0)))
        Eps.assign(dolfin.Constant(macro_strain))

        X = dolfin.SpatialCoordinate(self.mesh)
        X_0 = [0.]*self.dim
        for k_dim in range(self.dim):
            X_0[k_dim] = dolfin.assemble(X[k_dim] * dV)/mesh_V0
        X_0 = dolfin.Constant(X_0)

        u_bar   = dolfin.dot(Eps, X-X_0)
        u_tot = u_bar + v

        ########################################################### Solver ###############

        pf = dolfin.Constant(p_f)
        N = dolfin.FacetNormal(self.mesh)

        F = dolfin.inner(sigma(dv, 0, Eps), eps(v_))*dV - dolfin.inner(-pf * N, v_) * dS(0)

        a, L = dolfin.lhs(F), dolfin.rhs(F)
        a += dolfin.dot(lamb_,dv)*dV + dolfin.dot(dlamb,v_)*dV

        if (self.dim==3):
            dolfin.solve(a == L, w, [], solver_parameters={"linear_solver": "cg"}) 
        else:
            dolfin.solve(a == L, w, [], solver_parameters={"linear_solver": "lu"})  #MMT20230616: "cg" solver doesn't work when material Poisson ratio nu=0.499, and "lu" doesn't work for 3D geometries

        Vs0 = mesh_V0
        vs = dolfin.assemble((1 + dolfin.tr(eps(v)))*dV)
        V0 = vol
        Phi_s0 = Vs0/V0
        Phi_s = vs/V0
        kappa_tilde = Phi_s0**2 * p_f/(Phi_s0 - Phi_s)/2

        return kappa_tilde



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import collections
import dolfin

import dolfin_mech as dmech

################################################################################

class Problem():



    def __init__(self):

        self.subsols = collections.OrderedDict()

        self.operators = []
        self.constraints = []

        self.inelastic_behaviors_mixed    = []
        self.inelastic_behaviors_internal = []

        self.steps = []

        self.fois = []
        self.qois = []

        self.form_compiler_parameters = {}

####################################################################### mesh ###

    def set_mesh(self,
            mesh,
            define_spatial_coordinates=True,
            define_facet_normals=False,
            compute_bbox=False,
            compute_local_cylindrical_basis=False):

        self.dim = mesh.ufl_domain().geometric_dimension()

        self.mesh = mesh
        self.dV = dolfin.Measure(
            "dx",
            domain=self.mesh)
        self.mesh_V0 = dolfin.assemble(dolfin.Constant(1) * self.dV)

        if (define_spatial_coordinates):
            if "Inverse" in str(self):
                self.x = dolfin.SpatialCoordinate(self.mesh)
            else:
                self.X = dolfin.SpatialCoordinate(self.mesh)

        if (define_facet_normals):
            self.mesh_normals = dolfin.FacetNormal(mesh)

        if (compute_bbox):
            coord = self.mesh.coordinates()
            self.mesh_bbox = []
            xmin = dolfin.MPI.min(mesh.mpi_comm(), min(coord[:,0]))
            xmax = dolfin.MPI.max(mesh.mpi_comm(), max(coord[:,0]))
            self.mesh_bbox += [xmin, xmax]
            if (self.dim >= 2):
                ymin = dolfin.MPI.min(mesh.mpi_comm(), min(coord[:,1]))
                ymax = dolfin.MPI.max(mesh.mpi_comm(), max(coord[:,1]))
                self.mesh_bbox += [ymin, ymax]
                if (self.dim >= 3):
                    zmin = dolfin.MPI.min(mesh.mpi_comm(), min(coord[:,2]))
                    zmax = dolfin.MPI.max(mesh.mpi_comm(), max(coord[:,2]))
                    self.mesh_bbox += [zmin, zmax]

        if (compute_local_cylindrical_basis):
            self.local_basis_fe = dolfin.VectorElement(
                family="DG", # MG20220424: Why not CG?
                cell=mesh.ufl_cell(),
                degree=1)

            self.eR_expr = dolfin.Expression(
                ("+x[0]/sqrt(pow(x[0],2)+pow(x[1],2))", "+x[1]/sqrt(pow(x[0],2)+pow(x[1],2))"),
                element=self.local_basis_fe)
            self.eT_expr = dolfin.Expression(
                ("-x[1]/sqrt(pow(x[0],2)+pow(x[1],2))", "+x[0]/sqrt(pow(x[0],2)+pow(x[1],2))"),
                element=self.local_basis_fe)

            self.Q_expr = dolfin.as_matrix([[self.eR_expr[0], self.eR_expr[1]],
                                            [self.eT_expr[0], self.eT_expr[1]]])

            self.local_basis_fs = dolfin.FunctionSpace(
                mesh,
                self.local_basis_fe) # MG: element keyword don't work here…

            self.eR_func = dolfin.interpolate(
                v=self.eR_expr,
                V=self.local_basis_fs)
            self.eR_func.rename("eR", "eR")

            self.eT_func = dolfin.interpolate(
                v=self.eT_expr,
                V=self.local_basis_fs)
            self.eT_func.rename("eT", "eT")
        else:
            self.Q_expr = None



    def set_measures(self,
            domains=None,
            boundaries=None,
            points=None):

        self.domains = domains
        self.dV = dolfin.Measure(
            "cell",
            domain=self.mesh,
            subdomain_data=self.domains)
        # if (domains is not None):
        #     self.dV = dolfin.Measure(
        #         "dx",
        #         domain=self.mesh,
        #         subdomain_data=self.domains)
        # else:
        #     self.dV = dolfin.Measure(
        #         "dx",
        #         domain=self.mesh)

        self.boundaries = boundaries
        self.dS = dolfin.Measure(
            "exterior_facet",
            domain=self.mesh,
            subdomain_data=self.boundaries)
        # if (boundaries is not None):
        #     self.dS = dolfin.Measure(
        #         "ds",
        #         domain=self.mesh,
        #         subdomain_data=self.boundaries)
        # else:
        #     self.dS = dolfin.Measure(
        #         "ds",
        #         domain=self.mesh)

        self.points = points
        self.dP = dolfin.Measure(
            "vertex",
            domain=self.mesh,
            subdomain_data=self.points)
        # if (points is not None):
        #     self.dP = dolfin.Measure(
        #         "dP",
        #         domain=self.mesh,
        #         subdomain_data=self.points)
        # else:
        #     self.dP = dolfin.Measure(
        #         "dP",
        #         domain=self.mesh)

################################################################### solution ###

    def add_subsol(self,
            name,
            *args,
            **kwargs):

        subsol = dmech.SubSol(
            name=name,
            *args,
            **kwargs)
        self.subsols[name] = subsol
        return subsol



    def add_scalar_subsol(self,
            name,
            family="CG",
            degree=1,
            init_val=None,
            init_fun=None):

        fe = dolfin.FiniteElement(
            family=family,
            cell=self.mesh.ufl_cell(),
            degree=degree)

        self.add_subsol(
            name=name,
            fe=fe,
            init_val=init_val,
            init_fun=init_fun)



    def add_vector_subsol(self,
            name,
            family="CG",
            degree=1,
            init_val=None):

        fe = dolfin.VectorElement(
            family=family,
            cell=self.mesh.ufl_cell(),
            degree=degree)

        self.add_subsol(
            name=name,
            fe=fe,
            init_val=init_val)



    def add_tensor_subsol(self,
            name,
            family="CG",
            degree=1,
            symmetry=None,
            init_val=None):

        fe = dolfin.TensorElement(
            family=family,
            cell=self.mesh.ufl_cell(),
            degree=degree,
            symmetry=symmetry)

        self.add_subsol(
            name=name,
            fe=fe,
            init_val=init_val)



    def get_subsol(self,
            name):

        return self.subsols[name]



    def set_solution_finite_element(self):

        if (len(self.subsols) == 1):
            self.sol_fe = list(self.subsols.values())[0].fe
        else:
            self.sol_fe = dolfin.MixedElement([subsol.fe for subsol in self.subsols.values()])
        # print(self.sol_fe)



    def set_solution_function_space(self,
            constrained_domain=None):

        self.sol_fs = dolfin.FunctionSpace(
            self.mesh,
            self.sol_fe,
            constrained_domain=constrained_domain) # MG: element keyword don't work here…



    def get_subsol_function_space(self,
            name):

        index = list(self.subsols.keys()).index(name)
        # print(str(name)+" index = "+str(index))
        return self.sol_fs.sub(index)



    def set_solution_functions(self):

        self.sol_func     = dolfin.Function(self.sol_fs)
        self.sol_old_func = dolfin.Function(self.sol_fs)
        self.dsol_func    = dolfin.Function(self.sol_fs)
        self.dsol_test    = dolfin.TestFunction(self.sol_fs)
        self.dsol_tria    = dolfin.TrialFunction(self.sol_fs)

        if (len(self.subsols) == 1):
            subfuncs  = (self.sol_func,)
            dsubtests = (self.dsol_test,)
            dsubtrias = (self.dsol_tria,)
            funcs     = (self.sol_func,)
            funcs_old = (self.sol_old_func,)
            dfuncs    = (self.dsol_func,)
        else:
            subfuncs  = dolfin.split(self.sol_func)
            dsubtests = dolfin.split(self.dsol_test)
            dsubtrias = dolfin.split(self.dsol_tria)
            funcs     = dolfin.Function(self.sol_fs).split(deepcopy=1)
            funcs_old = dolfin.Function(self.sol_fs).split(deepcopy=1)
            dfuncs    = dolfin.Function(self.sol_fs).split(deepcopy=1)

        for (k_subsol,subsol) in enumerate(self.subsols.values()):
            subsol.subfunc  = subfuncs[k_subsol]
            subsol.dsubtest = dsubtests[k_subsol]
            subsol.dsubtria = dsubtrias[k_subsol]

            subsol.func = funcs[k_subsol]
            subsol.func.rename(subsol.name, subsol.name)
            subsol.func_old = funcs_old[k_subsol]
            subsol.func_old.rename(subsol.name+"_old", subsol.name+"_old")
            subsol.dfunc = dfuncs[k_subsol]
            subsol.dfunc.rename("d"+subsol.name, "d"+subsol.name)

        for (k_subsol,subsol) in enumerate(self.subsols.values()):
            subsol.init()
        if (len(self.subsols) > 1):
            dolfin.assign(
                self.sol_func,
                self.get_subsols_func_lst())
            dolfin.assign(
                self.sol_old_func,
                self.get_subsols_func_old_lst())



    def get_subsols_func_lst(self):

        return [subsol.func for subsol in self.subsols.values()]



    def get_subsols_func_old_lst(self):

        return [subsol.func_old for subsol in self.subsols.values()]



    def get_subsols_dfunc_lst(self):

        return [subsol.dfunc for subsol in self.subsols.values()]



    def set_quadrature_degree(self,
            quadrature_degree):

        self.form_compiler_parameters["quadrature_degree"] = quadrature_degree

######################################################################## FOI ###

    def set_foi_finite_elements_DG(self,
            degree=0): # MG20180420: DG elements are simpler to manage than quadrature elements, since quadrature elements must be compatible with the expression's degree, which is not always trivial (e.g., for J…)

        self.sfoi_fe = dolfin.FiniteElement(
            family="DG",
            cell=self.mesh.ufl_cell(),
            degree=degree)

        self.vfoi_fe = dolfin.VectorElement(
            family="DG",
            cell=self.mesh.ufl_cell(),
            degree=degree)

        self.mfoi_fe = dolfin.TensorElement(
            family="DG",
            cell=self.mesh.ufl_cell(),
            degree=degree)



    def set_foi_finite_elements_Quad(self,
            degree=0): # MG20180420: DG elements are simpler to manage than quadrature elements, since quadrature elements must be compatible with the expression's degree, which is not always trivial (e.g., for J…)

        self.sfoi_fe = dolfin.FiniteElement(
            family="Quadrature",
            cell=self.mesh.ufl_cell(),
            degree=degree,
            quad_scheme="default")
        self.sfoi_fe._quad_scheme = "default"           # MG20180406: is that even needed?
        for sub_element in self.sfoi_fe.sub_elements(): # MG20180406: is that even needed?
            sub_element._quad_scheme = "default"        # MG20180406: is that even needed?

        self.vfoi_fe = dolfin.VectorElement(
            family="Quadrature",
            cell=self.mesh.ufl_cell(),
            degree=degree,
            quad_scheme="default")
        self.vfoi_fe._quad_scheme = "default"           # MG20180406: is that even needed?
        for sub_element in self.vfoi_fe.sub_elements(): # MG20180406: is that even needed?
            sub_element._quad_scheme = "default"        # MG20180406: is that even needed?

        self.mfoi_fe = dolfin.TensorElement(
            family="Quadrature",
            cell=self.mesh.ufl_cell(),
            degree=degree,
            quad_scheme="default")
        self.mfoi_fe._quad_scheme = "default"           # MG20180406: is that still needed?
        for sub_element in self.mfoi_fe.sub_elements(): # MG20180406: is that still needed?
            sub_element._quad_scheme = "default"        # MG20180406: is that still needed?



    def set_foi_function_spaces(self):

        self.sfoi_fs = dolfin.FunctionSpace(
            self.mesh,
            self.sfoi_fe) # MG: element keyword don't work here…

        self.vfoi_fs = dolfin.FunctionSpace(
            self.mesh,
            self.vfoi_fe) # MG: element keyword don't work here…

        self.mfoi_fs = dolfin.FunctionSpace(
            self.mesh,
            self.mfoi_fe) # MG: element keyword don't work here…



    def add_foi(self, *args, **kwargs):

        foi = dmech.FOI(
            *args,
            form_compiler_parameters=self.form_compiler_parameters,
            **kwargs)
        self.fois += [foi]
        return foi



    def get_foi(self, name):

        for foi in self.fois:
            if (foi.name == name): return foi
        assert (0),\
            "No FOI named \""+name+"\". Aborting."



    def update_fois(self):

        for foi in self.fois:
            foi.update()



    def get_fois_func_lst(self):

        return [foi.func for foi in self.fois]

######################################################################## QOI ###

    def add_qoi(self, *args, **kwargs):

        qoi = dmech.QOI(
            *args,
            form_compiler_parameters=self.form_compiler_parameters,
            **kwargs)
        self.qois += [qoi]
        return qoi



    def update_qois(self, dt=None):

        for qoi in self.qois:
            qoi.update(dt)

################################################################## operators ###

    def add_operator(self,
            operator,
            k_step=None):

        if (k_step is None):
            self.operators += [operator]
        else:
            self.steps[k_step].operators += [operator]
        return operator

################################################################## operators ###

# MG20230131: Loading operators should not be there,
# but they are shared between Elasticity & HyperElasticity problems,
# so it is more convenient for the moment.

    def add_volume_force0_loading_operator(self,
            k_step=None,
            **kwargs):

        operator = dmech.VolumeForce0LoadingOperator(
            U_test=self.get_displacement_subsol().dsubtest,
            **kwargs)
        return self.add_operator(operator=operator, k_step=k_step)



    def add_volume_force_loading_operator(self,
            k_step=None,
            **kwargs):

        operator = dmech.VolumeForceLoadingOperator(
            U_test=self.get_displacement_subsol().dsubtest,
            kinematics=self.kinematics,
            **kwargs)
        return self.add_operator(operator=operator, k_step=k_step)



    def add_surface_force0_loading_operator(self,
            k_step=None,
            **kwargs):

        operator = dmech.SurfaceForce0LoadingOperator(
            U_test=self.get_displacement_subsol().dsubtest,
            **kwargs)
        return self.add_operator(operator=operator, k_step=k_step)



    def add_surface_force_loading_operator(self,
            k_step=None,
            **kwargs):

        operator = dmech.SurfaceForceLoadingOperator(
            U_test=self.get_displacement_subsol().dsubtest,
            kinematics=self.kinematics,
            N=self.mesh_normals,
            **kwargs)
        return self.add_operator(operator=operator, k_step=k_step)



    def add_surface_pressure0_loading_operator(self,
            k_step=None,
            **kwargs):

        operator = dmech.SurfacePressure0LoadingOperator(
            U_test=self.get_displacement_subsol().dsubtest,
            N=self.mesh_normals,
            **kwargs)
        return self.add_operator(operator=operator, k_step=k_step)



    def add_surface_pressure_loading_operator(self,
            k_step=None,
            **kwargs):

        operator = dmech.SurfacePressureLoadingOperator(
            U_test=self.get_displacement_subsol().dsubtest,
            kinematics=self.kinematics,
            N=self.mesh_normals,
            **kwargs)
        return self.add_operator(operator=operator, k_step=k_step)



    def add_surface_pressure_gradient0_loading_operator(self,
            k_step=None,
            **kwargs):

        operator = dmech.SurfacePressureGradient0LoadingOperator(
            x=dolfin.SpatialCoordinate(self.mesh),
            U_test=self.get_displacement_subsol().dsubtest,
            N=self.mesh_normals,
            **kwargs)
        return self.add_operator(operator=operator, k_step=k_step)



    def add_surface_pressure_gradient_loading_operator(self,
            k_step=None,
            **kwargs):

        operator = dmech.SurfacePressureGradientLoadingOperator(
            X=dolfin.SpatialCoordinate(self.mesh),
            U=self.get_displacement_subsol().subfunc,
            U_test=self.get_displacement_subsol().dsubtest,
            kinematics=self.kinematics,
            N=self.mesh_normals,
            **kwargs)
        return self.add_operator(operator=operator, k_step=k_step)



    def add_surface_tension0_loading_operator(self,
            k_step=None,
            **kwargs):

        operator = dmech.SurfaceTension0LoadingOperator(
            u=self.get_displacement_subsol().subfunc,
            u_test=self.get_displacement_subsol().dsubtest,
            kinematics=self.kinematics,
            N=self.mesh_normals,
            **kwargs)
        return self.add_operator(operator=operator, k_step=k_step)



    def add_surface_tension_loading_operator(self,
            k_step=None,
            **kwargs):

        operator = dmech.SurfaceTensionLoadingOperator(
            U=self.get_displacement_subsol().subfunc,
            U_test=self.get_displacement_subsol().dsubtest,
            kinematics=self.kinematics,
            N=self.mesh_normals,
            **kwargs)
        return self.add_operator(operator=operator, k_step=k_step)



    def add_normal_displacement_penalty_operator(self,
            k_step=None,
            **kwargs):

        operator = dmech.NormalDisplacementPenaltyOperator(
            U=self.get_displacement_subsol().subfunc,
            U_test=self.get_displacement_subsol().dsubtest,
            N=self.mesh_normals,
            **kwargs)
        return self.add_operator(operator=operator, k_step=k_step)



    def add_directional_displacement_penalty_operator(self,
            k_step=None,
            **kwargs):

        operator = dmech.DirectionalDisplacementPenaltyOperator(
            U=self.get_displacement_subsol().subfunc,
            U_test=self.get_displacement_subsol().dsubtest,
            **kwargs)
        return self.add_operator(operator=operator, k_step=k_step)



    def add_inertia_operator(self,
            k_step=None,
            **kwargs):

        operator = dmech.InertiaOperator(
            U=self.get_displacement_subsol().subfunc,
            U_old=self.get_displacement_subsol().func_old,
            U_test=self.get_displacement_subsol().dsubtest,
            **kwargs)
        return self.add_operator(operator=operator, k_step=k_step)

################################################################ constraints ###

    def add_constraint(self,
            *args,
            k_step=None,
            **kwargs):

        constraint = dmech.Constraint(*args, **kwargs)
        if (k_step is None):
            self.constraints += [constraint]
        else:
            self.steps[k_step].constraints += [constraint]
        return constraint

###################################################################### steps ###

    def add_step(self,
            Deltat=1.,
            **kwargs):

        if len(self.steps) == 0:
            t_ini = 0.
            t_fin = Deltat
        else:
            t_ini = self.steps[-1].t_fin
            t_fin = t_ini + Deltat
        step = dmech.Step(
            t_ini=t_ini,
            t_fin=t_fin,
            **kwargs)
        self.steps += [step]
        return len(self.steps)-1

###################################################################### forms ###

    def set_variational_formulation(self,
            k_step=None):

        self.res_form = sum([operator.res_form for operator in self.operators if (operator.measure.integral_type() != "vertex")]) # MG20190513: Cannot use point integral within assemble_system
        if (k_step is not None):
            self.res_form += sum([operator.res_form for operator in self.steps[k_step].operators if (operator.measure.integral_type() != "vertex")]) # MG20190513: Cannot use point integral within assemble_system

        # print(self.res_form)
        # for operator in self.operators:
        #     if (operator.measure.integral_type() != "vertex"):
        #         print(type(operator))
        #         print(operator.res_form)

        self.jac_form = dolfin.derivative(
            self.res_form,
            self.sol_func,
            self.dsol_tria)

        # print(self.jac_form)



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin

import dolfin_mech as dmech
from .Operator import Operator

################################################################################

class PfPoroOperator(Operator):

    def __init__(self,
            Phis_test,
            measure,
            pf_val=None, pf_ini=None, pf_fin=None):

        self.measure = measure

        self.tv_pf = dmech.TimeVaryingConstant(
            val=pf_val, val_ini=pf_ini, val_fin=pf_fin)
        self.pf = self.tv_pf.val

        self.res_form = dolfin.inner(self.pf, Phis_test) * self.measure



    def set_value_at_t_step(self,
            t_step):

        self.tv_pf.set_value_at_t_step(t_step)



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
###                                                                          ###
### And Cécile Patte, 2019-2021                                              ###
###                                                                          ###
### INRIA, Palaiseau, France                                                 ###
###                                                                          ###
################################################################################

import dolfin

import dolfin_mech as dmech
from .Material_Elastic import ElasticMaterial

################################################################################

class ExponentialOgdenCiarletGeymonatElasticMaterial(ElasticMaterial):



    def __init__(self,
            kinematics,
            parameters):

        self.kinematics = kinematics

        if ("alpha" in parameters) and ("gamma" in parameters):
            self.alpha = dolfin.Constant(parameters["alpha"])
            self.gamma = dolfin.Constant(parameters["gamma"])
        else:
            assert (0),\
                "No parameter found: \"+str(parameters)+\". Must provide alpha & gamma. Aborting."

        self.Psi = (self.alpha) * (dolfin.exp(self.gamma*(self.kinematics.J**2 - 1 - 2*dolfin.ln(self.kinematics.J))) - 1)

        self.Sigma = (self.alpha) * dolfin.exp(self.gamma*(self.kinematics.J**2 - 1 - 2*dolfin.ln(self.kinematics.J))) * (2*self.gamma) * (self.kinematics.J**2 - 1) * self.kinematics.C_inv

        # self.P = dolfin.diff(self.Psi, self.kinematics.F) # MG20220426: Cannot do that for micromechanics problems
        self.P = self.kinematics.F * self.Sigma

        self.sigma = self.P * self.kinematics.F.T / self.kinematics.J



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin

import dolfin_mech as dmech
from .Operator import Operator

################################################################################

class InertiaOperator(Operator):

    def __init__(self,
            U,
            U_old,
            U_test,
            measure,
            rho_val=None, rho_ini=None, rho_fin=None):

        self.measure = measure

        self.tv_rho = dmech.TimeVaryingConstant(
            val=rho_val, val_ini=rho_ini, val_fin=rho_fin)
        rho = self.tv_rho.val

        self.tv_dt = dmech.TimeVaryingConstant(0.)
        dt = self.tv_dt.val

        # Pi = (rho/2/dt) * dolfin.inner(U, U)**2 * self.measure # MG20221108: What was that?!

        V = (U - U_old)/dt
        Pi = (rho/2) * dolfin.inner(V, V) * self.measure
        self.res_form = dolfin.derivative(Pi, U, U_test)



    def set_value_at_t_step(self,
            t_step):

        self.tv_rho.set_value_at_t_step(t_step)



    def set_dt(self,
            dt):

        self.tv_dt.set_value(dt)



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin

import dolfin_mech as dmech
from .Material_Elastic import ElasticMaterial

################################################################################

class NeoHookeanElasticMaterial(ElasticMaterial):



    def __init__(self,
            kinematics,
            parameters,
            decoup=False):

        self.kinematics = kinematics

        self.C1 = self.get_C1_from_parameters(parameters)

        if (decoup):
            if   (self.kinematics.dim == 2):
                # self.Psi      =   self.C1 * (self.kinematics.J**(-2/3) * (self.kinematics.J * self.kinematics.IC_bar + 1) - 3) # MG20200206: Plane strain, written with IC_bar_2D (I prefer the expression with IC_2D)
                self.Psi      =   self.C1 * (self.kinematics.J**(-2/3) * (self.kinematics.IC + 1) - 3) # MG20200206: Plane strain, written with IC_2D
                self.Sigma    = 2*self.C1 * self.kinematics.J**(-2/3) * (self.kinematics.I - (self.kinematics.IC + 1)/3 * self.kinematics.C_inv) # MG20200206: Cannot differentiate Psi wrt to C because J is not defined as a function of C
                self.Sigma_ZZ = 2*self.C1 * self.kinematics.J**(-2/3) * (1 - (self.kinematics.IC + 1)/3)
            elif (self.kinematics.dim == 3):
                self.Psi   =   self.C1 * (self.kinematics.IC_bar - 3)
                self.Sigma = 2*self.C1 * self.kinematics.J**(-2/3) * (self.kinematics.I - self.kinematics.IC/3 * self.kinematics.C_inv) # MG20200206: Cannot differentiate Psi wrt to C because J is not defined as a function of C
        else:
            if   (self.kinematics.dim == 2):
                self.Psi      =   self.C1 * (self.kinematics.IC - 2 - 2*dolfin.ln(self.kinematics.J)) # MG20200206: Plane strain
                self.Sigma    = 2*self.C1 * (self.kinematics.I - self.kinematics.C_inv) # MG20200206: Cannot differentiate Psi wrt to C because J is not defined as a function of C
                self.Sigma_ZZ = dolfin.Constant(0.)
            elif (self.kinematics.dim == 3):
                self.Psi   =   self.C1 * (self.kinematics.IC - 3 - 2*dolfin.ln(self.kinematics.J))
                self.Sigma = 2*self.C1 * (self.kinematics.I - self.kinematics.C_inv) # MG20200206: Cannot differentiate Psi wrt to C because J is not defined as a function of C

        # self.P = dolfin.diff(self.Psi, self.kinematics.F) # MG20220426: Cannot do that for micromechanics problems
        self.P = self.kinematics.F * self.Sigma

        self.sigma = self.P * self.kinematics.F.T / self.kinematics.J



    # def get_free_energy(self,
    #         U=None,
    #         C=None):

    #     C     = self.get_C_from_U_or_C(U,C)
    #     IC    = dolfin.tr(C)
    #     JF    = dolfin.sqrt(dolfin.det(C)) # MG20200207: Watch out! This is well defined for inverted elements!
    #     # C_inv = dolfin.inv(C)

    #     assert (C.ufl_shape[0] == C.ufl_shape[1])
    #     dim = C.ufl_shape[0]
    #     # I = dolfin.Identity(dim)

    #     if   (dim == 2):
    #         Psi   =   self.C1 * (IC - 2 - 2*dolfin.ln(JF)) # MG20200206: plane strain
    #         # Sigma = 2*self.C1 * (I - C_inv)
    #     elif (dim == 3):
    #         Psi   =   self.C1 * (IC - 3 - 2*dolfin.ln(JF))
    #         # Sigma = 2*self.C1 * (I - C_inv)
    #     Sigma = 2*dolfin.diff(Psi, C)

    #     return Psi, Sigma



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
###                                                                          ###
### And Mahdi Manoochehrtayebi, 2020-2024                                    ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin

import dolfin_mech as dmech
from .Operator import Operator

################################################################################

class MacroscopicStretchComponentPenaltyOperator(Operator):

    def __init__(self,
            U_bar,
            U_bar_test,
            i, j,
            measure,
            U_bar_ij_val=None, U_bar_ij_ini=None, U_bar_ij_fin=None,
            pen_val=None, pen_ini=None, pen_fin=None):

        self.measure = measure

        self.tv_U_bar_ij = dmech.TimeVaryingConstant(
            val=U_bar_ij_val, val_ini=U_bar_ij_ini, val_fin=U_bar_ij_fin)
        U_bar_ij = self.tv_U_bar_ij.val

        self.tv_pen = dmech.TimeVaryingConstant(
            val=pen_val, val_ini=pen_ini, val_fin=pen_fin)
        pen = self.tv_pen.val

        Pi = (pen/2) * (U_bar[i,j] - U_bar_ij)**2 * self.measure
        self.res_form = dolfin.derivative(Pi, U_bar[i,j], U_bar_test[i,j])



    def set_value_at_t_step(self,
            t_step):

        self.tv_U_bar_ij.set_value_at_t_step(t_step)
        self.tv_pen.set_value_at_t_step(t_step)



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin
import numpy

import dolfin_mech as dmech
from .Problem_Hyperelasticity import HyperelasticityProblem

################################################################################

class PoroHyperelasticityProblem(HyperelasticityProblem):



    def __init__(self,
            mesh=None,
            define_facet_normals=False,
            domains_mf=None,
            boundaries_mf=None,
            points_mf=None,
            displacement_degree=1,
            porosity_degree=None,
            quadrature_degree=None,
            foi_degree=0,
            porosity_init_val=None,
            porosity_init_fun=None,
            skel_behavior=None,
            skel_behaviors=[],
            bulk_behavior=None,
            bulk_behaviors=[],
            pore_behavior=None,
            pore_behaviors=[],
            gradient_operators=None):

        HyperelasticityProblem.__init__(self)

        if (mesh is not None):
            self.set_mesh(
                mesh=mesh,
                define_facet_normals=define_facet_normals)

            self.set_measures(
                domains=domains_mf,
                boundaries=boundaries_mf,
                points=points_mf)

            self.set_subsols(
                displacement_degree=displacement_degree,
                porosity_degree=porosity_degree,
                porosity_init_val=porosity_init_val,
                porosity_init_fun=porosity_init_fun)
            
            assert (porosity_init_val is None) or (porosity_init_fun is None)
            if gradient_operators=="direct":
                self.Phis0 = porosity_init_val if (porosity_init_fun is None) else porosity_init_fun ### AP2023 - defined here to initialise subsol center of gravity
                self.set_subsols_gradient_direct()
            elif gradient_operators=="inverse":
                self.set_subsols_gradient_inverse()
            
            self.set_solution_finite_element()
            self.set_solution_function_space()
            self.set_solution_functions()

            self.set_quadrature_degree(
                quadrature_degree=quadrature_degree)

            self.set_foi_finite_elements_DG(
                degree=foi_degree)
            self.set_foi_function_spaces()

            self.set_kinematics()

            assert (porosity_init_val is None) or (porosity_init_fun is None)
            self.init_known_porosity(
                porosity_init_val=porosity_init_val,
                porosity_init_fun=porosity_init_fun)

            assert (skel_behavior is     None) or (len(skel_behaviors)==0),\
                "Cannot provide both skel_behavior & skel_behaviors. Aborting."
            assert (skel_behavior is not None) or (len(skel_behaviors) >0),\
                "Need to provide skel_behavior or skel_behaviors. Aborting."
            if (skel_behavior is not None):
                skel_behaviors = [skel_behavior]
            self.add_Wskel_operators(skel_behaviors)

            assert (bulk_behavior is     None) or (len(bulk_behaviors)==0),\
                "Cannot provide both bulk_behavior & bulk_behaviors. Aborting."
            assert (bulk_behavior is not None) or (len(bulk_behaviors) >0),\
                "Need to provide bulk_behavior or bulk_behaviors. Aborting."
            if (bulk_behavior is not None):
                bulk_behaviors = [bulk_behavior]
            self.add_Wbulk_operators(bulk_behaviors)

            assert (pore_behavior is None) or (len(pore_behaviors)==0),\
                "Cannot provide both pore_behavior & pore_behaviors. Aborting."
            if (pore_behavior is not None):
                pore_behaviors = [pore_behavior]
            self.add_Wpore_operators(pore_behaviors)



    def get_porosity_name(self):
        return "Phis"



    def add_porosity_subsol(self,
            degree,
            init_val=None,
            init_fun=None):

        if (degree == 0):
            self.add_scalar_subsol(
                name=self.get_porosity_name(),
                family="DG",
                degree=0,
                init_val=init_val,
                init_fun=init_fun)
        else:
            self.add_scalar_subsol(
                name=self.get_porosity_name(),
                family="CG",
                degree=degree,
                init_val=init_val,
                init_fun=init_fun)



    def get_porosity_subsol(self):

        return self.get_subsol(self.get_porosity_name())

    def get_porosity_function_space(self):

        return self.get_subsol_function_space(name=self.get_porosity_name())
    


    def get_balanced_gravity_boundary_pressure_name(self):
            return "pressure_boundary_condition"

    def add_balanced_gravity_boundary_pressure_subsol(self):

        self.add_scalar_subsol(
            name=self.get_balanced_gravity_boundary_pressure_name(),
            family="CG",
            degree=1)
        
    def get_balanced_gravity_boundary_pressure_subsol(self):

        return self.get_subsol(self.get_balanced_gravity_boundary_pressure_name())

    def get_balanced_gravity_boundary_pressure_function_space(self):

        return self.get_subsol_function_space(name=self.get_balanced_gravity_boundary_pressure_name())
    


    def get_lbda_name(self):

        return "lbda"

    def add_lbda_subsol(self,
            init_val=None):

        self.add_vector_subsol(
            name=self.get_lbda_name(),
            family="R",
            degree=0,
            init_val=init_val)
    
    def get_lbda_subsol(self):

        return self.get_subsol(self.get_lbda_name())



    def get_mu_name(self):

        return "mu"
    
    def add_mu_subsol(self,
            init_val=None):

        self.add_vector_subsol(
            name=self.get_mu_name(),
            family="R",
            degree=0,
            init_val=init_val)
        
    def get_mu_subsol(self):

        return self.get_subsol(self.get_mu_name())
    

    
    def get_gamma_name(self):
        return "gamma"

    def add_gamma_subsol(self):

        self.add_scalar_subsol(
            name=self.get_gamma_name(),
            family="R",
            degree=0)
        
    def get_gamma_subsol(self):

        return self.get_subsol(self.get_gamma_name())

    def get_gamma_function_space(self):

        return self.get_subsol_function_space(name=self.get_gamma_name())
    


    def get_center_gravity_direct_problem_name(self):

        return "xg"
            
    def add_center_gravity_direct_problem_subsol(self,
            init_val=None):
        
        self.add_vector_subsol(
            name=self.get_center_gravity_direct_problem_name(),
            family="R",
            degree=0,
            init_val=self.get_center_gravity())

    def get_center_gravity_direct_problem_subsol(self):
        return self.get_subsol(self.get_center_gravity_direct_problem_name())



    def set_subsols(self,
            displacement_degree=1,
            porosity_degree=None,
            porosity_init_val=None,
            porosity_init_fun=None):

        self.add_displacement_subsol(
            degree=displacement_degree)

        if (porosity_degree is None):
            porosity_degree = displacement_degree-1
        self.add_porosity_subsol(
            degree=porosity_degree,
            init_val=porosity_init_val,
            init_fun=porosity_init_fun)



    def set_subsols_gradient_direct(self):
            
                self.add_balanced_gravity_boundary_pressure_subsol()
                self.add_gamma_subsol()
                self.add_lbda_subsol()
                self.add_mu_subsol()
                self.add_center_gravity_direct_problem_subsol()



    def set_subsols_gradient_inverse(self):
        
            self.add_balanced_gravity_boundary_pressure_subsol()
            self.add_gamma_subsol()
            self.add_lbda_subsol()
            self.add_mu_subsol()



    def init_known_porosity(self,
            porosity_init_val,
            porosity_init_fun):

        if   (porosity_init_val   is not None):
            self.Phis0 = dolfin.Constant(porosity_init_val)
        elif (porosity_init_fun is not None):
            self.Phis0 = porosity_init_fun
        self.add_foi(
            expr=self.Phis0,
            fs=self.get_porosity_function_space().collapse(),
            name="Phis0")
        self.add_foi(
            expr=1 - self.Phis0,
            fs=self.get_porosity_function_space().collapse(),
            name="Phif0")
        self.add_foi(
            expr=self.kinematics.J - self.get_porosity_subsol().subfunc,
            fs=self.get_porosity_function_space().collapse(),
            name="Phif")
        self.add_foi(
            expr=self.get_porosity_subsol().subfunc/self.kinematics.J,
            fs=self.get_porosity_function_space().collapse(),
            name="phis")
        self.add_foi(
            expr=1.-self.get_porosity_subsol().subfunc/self.kinematics.J,
            fs=self.get_porosity_function_space().collapse(),
            name="phif")



    def add_Wskel_operator(self,
            material_parameters,
            material_scaling,
            subdomain_id=None):

        operator = dmech.WskelPoroOperator(
            kinematics=self.kinematics,
            U_test=self.get_displacement_subsol().dsubtest,
            Phis0=self.Phis0,
            material_parameters=material_parameters,
            material_scaling=material_scaling,
            measure=self.get_subdomain_measure(subdomain_id))
        return self.add_operator(operator)



    def add_Wskel_operators(self,
            skel_behaviors):

        for skel_behavior in skel_behaviors:
            operator = self.add_Wskel_operator(
                material_parameters=skel_behavior["parameters"],
                material_scaling=skel_behavior["scaling"],
                subdomain_id=skel_behavior.get("subdomain_id", None))
            suffix = "_"+skel_behavior["suffix"] if "suffix" in skel_behavior else ""
            self.add_foi(expr=operator.material.Sigma, fs=self.mfoi_fs, name="Sigma_skel"+suffix)
            self.add_foi(expr=operator.material.sigma, fs=self.mfoi_fs, name="sigma_skel"+suffix)



    def add_Wbulk_operator(self,
            material_parameters,
            material_scaling,
            subdomain_id=None):

        operator = dmech.WbulkPoroOperator(
            kinematics=self.kinematics,
            U_test=self.get_displacement_subsol().dsubtest,
            Phis0=self.Phis0,
            Phis=self.get_porosity_subsol().subfunc,
            Phis_test=self.get_porosity_subsol().dsubtest,
            material_parameters=material_parameters,
            material_scaling=material_scaling,
            measure=self.get_subdomain_measure(subdomain_id))
        return self.add_operator(operator)



    def add_Wbulk_operators(self,
            bulk_behaviors):

        for bulk_behavior in bulk_behaviors:
            operator = self.add_Wbulk_operator(
                material_parameters=bulk_behavior["parameters"],
                material_scaling=bulk_behavior["scaling"],
                subdomain_id=bulk_behavior.get("subdomain_id", None))
            suffix = "_"+bulk_behavior["suffix"] if "suffix" in bulk_behavior else ""
            self.add_foi(expr=operator.material.dWbulkdPhis, fs=self.sfoi_fs, name="dWbulkdPhis"+suffix)
            self.add_foi(expr=operator.material.dWbulkdPhis * self.kinematics.J * self.kinematics.C_inv, fs=self.mfoi_fs, name="Sigma_bulk"+suffix)
            self.add_foi(expr=operator.material.dWbulkdPhis * self.kinematics.I, fs=self.mfoi_fs, name="sigma_bulk"+suffix)



    def add_Wpore_operator(self,
            material_parameters,
            material_scaling,
            subdomain_id=None):

        operator = dmech.WporePoroOperator(
            kinematics=self.kinematics,
            Phis0=self.Phis0,
            Phis=self.get_porosity_subsol().subfunc,
            Phis_test=self.get_porosity_subsol().dsubtest,
            material_parameters=material_parameters,
            material_scaling=material_scaling,
            measure=self.get_subdomain_measure(subdomain_id))
        return self.add_operator(operator)



    def add_Wpore_operators(self,
            pore_behaviors):

        for pore_behavior in pore_behaviors:
            self.add_Wpore_operator(
                material_parameters=pore_behavior["parameters"],
                material_scaling=pore_behavior["scaling"],
                subdomain_id=pore_behavior.get("subdomain_id", None))



    def add_pf_operator(self,
            k_step=None,
            **kwargs):

        operator = dmech.PfPoroOperator(
            Phis_test=self.get_porosity_subsol().dsubtest,
            **kwargs)
        self.add_operator(
            operator=operator,
            k_step=k_step)
        self.add_foi(expr=operator.pf, fs=self.sfoi_fs, name="pf")



    def add_pressure_balancing_gravity0_loading_operator(self,
            k_step=None,
            **kwargs):

        operator = dmech.PressureBalancingGravity0LoadingOperator(
            x = self.x,
            x0 = self.get_center_gravity(),
            n = self.mesh_normals,
            u_test = self.get_displacement_subsol().dsubtest, 
            lbda = self.get_lbda_subsol().subfunc,
            lbda_test = self.get_lbda_subsol().dsubtest,
            p = self.get_balanced_gravity_boundary_pressure_subsol().subfunc,
            p_test = self.get_balanced_gravity_boundary_pressure_subsol().dsubtest,
            gamma = self.get_gamma_subsol().subfunc,
            gamma_test = self.get_gamma_subsol().dsubtest,
            mu = self.get_mu_subsol().subfunc,
            mu_test= self.get_mu_subsol().dsubtest,
            **kwargs)
        return self.add_operator(operator=operator, k_step=k_step)



    def add_pressure_balancing_gravity_loading_operator(self,
            k_step=None,
            **kwargs):

        operator = dmech.PressureBalancingGravityLoadingOperator(
            X=self.X,
            x0=self.get_center_gravity_direct_problem_subsol().subfunc,
            x0_test=self.get_center_gravity_direct_problem_subsol().dsubtest,
            lbda=self.get_lbda_subsol().subfunc,
            lbda_test=self.get_lbda_subsol().dsubtest,
            mu=self.get_mu_subsol().subfunc,
            mu_test=self.get_mu_subsol().dsubtest,
            p = self.get_balanced_gravity_boundary_pressure_subsol().subfunc,
            p_test = self.get_balanced_gravity_boundary_pressure_subsol().dsubtest,
            gamma = self.get_gamma_subsol().subfunc,
            gamma_test = self.get_gamma_subsol().dsubtest,
            kinematics=self.kinematics,
            U=self.get_displacement_subsol().subfunc,
            U_test=self.get_displacement_subsol().dsubtest,
            Phis0=self.Phis0,
            N=self.mesh_normals,
            **kwargs)
        return self.add_operator(operator=operator, k_step=k_step)



    def add_global_porosity_qois(self):

        self.add_qoi(
            name=self.get_porosity_name(),
            expr=self.get_porosity_subsol().subfunc * self.dV)

        self.add_qoi(
            name="Phif",
            expr=(self.kinematics.J - self.get_porosity_subsol().subfunc) * self.dV)
            
        self.add_qoi(
            name="phis",
            expr=(self.get_porosity_subsol().subfunc/self.kinematics.J) * self.dV)
            
        self.add_qoi(
            name="phif",
            expr=(1. - self.get_porosity_subsol().subfunc/self.kinematics.J) * self.dV)
        

    
    def get_center_gravity(self): 
        if "Inverse" in str(self):
            porosity = self.phis 
            position = self.x
        else:
            porosity=self.Phis0
            position=self.X
        center_gravity = numpy.empty(self.dim)
        for k_dim in range(self.dim):
            center_gravity[k_dim] = dolfin.assemble(porosity*position[k_dim]*self.dV)/dolfin.assemble(porosity*self.dV)
        return(center_gravity)



    def add_global_stress_qois(self,
            stress_type="cauchy"):

        if (stress_type in ("Cauchy", "cauchy", "sigma")):
            basename = "s_"
            stress = "sigma"
        elif (stress_type in ("Piola", "piola", "PK2", "Sigma")):
            basename = "S_"
            stress = "Sigma"
        elif (stress_type in ("Boussinesq", "boussinesq", "PK1", "P")):
            assert (0), "ToDo. Aborting."

        compnames = ["XX"]
        comps     = [(0,0)]
        if (self.dim >= 2):
            compnames += ["YY"]
            comps     += [(1,1)]
            if (self.dim >= 3):
                compnames += ["ZZ"]
                comps     += [(2,2)]
            compnames += ["XY"]
            comps     += [(0,1)]
            if (self.dim >= 3):
                compnames += ["YZ"]
                comps     += [(1,2)]
                compnames += ["ZX"]
                comps     += [(2,0)]
        for compname, comp in zip(compnames, comps):
            if (stress == "Sigma"):
                self.add_qoi(
                    name=basename+"skel_"+compname,
                    expr=sum([getattr(operator.material, stress)[comp]*operator.measure for operator in self.operators if (hasattr(operator, "material") and hasattr(operator.material, stress))]))
                self.add_qoi(
                    name=basename+"bulk_"+compname,
                    expr=sum([getattr(operator.material, "dWbulkdPhis")*self.kinematics.J*self.kinematics.C_inv[comp]*operator.measure for operator in self.operators if (hasattr(operator, "material") and hasattr(operator.material, "dWbulkdPhis"))]))
                self.add_qoi(
                    name=basename+"tot_"+compname,
                    expr=sum([getattr(operator.material, stress)[comp]*operator.measure for operator in self.operators if (hasattr(operator, "material") and hasattr(operator.material, stress))]))+sum([getattr(operator.material, "dWbulkdPhis")[comp]*self.kinematics.J*self.kinematics.C_inv*operator.measure for operator in self.operators if (hasattr(operator, "material") and hasattr(operator.material, "dWbulkdPhis"))])
            elif (stress == "sigma"):
                self.add_qoi(
                    name=basename+"skel_"+compname,
                    expr=sum([getattr(operator.material, stress)[comp]*self.kinematics.J*operator.measure for operator in self.operators if (hasattr(operator, "material") and hasattr(operator.material, stress))]))
                self.add_qoi(
                    name=basename+"bulk_"+compname,
                    expr=sum([getattr(operator.material, "dWbulkdPhis")*self.kinematics.I[comp]*self.kinematics.J*operator.measure for operator in self.operators if (hasattr(operator, "material") and hasattr(operator.material, "dWbulkdPhis"))]))
                self.add_qoi(
                    name=basename+"tot_"+compname,
                    expr=sum([getattr(operator.material, stress)[comp]*self.kinematics.J*operator.measure for operator in self.operators if (hasattr(operator, "material") and hasattr(operator.material, stress))])+sum([getattr(operator.material, "dWbulkdPhis")*self.kinematics.I[comp]*self.kinematics.J*operator.measure for operator in self.operators if (hasattr(operator, "material") and hasattr(operator.material, "dWbulkdPhis"))]))



    def add_global_fluid_pressure_qoi(self):

        # for operator in self.operators:
        #     print(type(operator))
        #     print(hasattr(operator, "pf"))

        # for step in self.steps:
        #     print(step)
        #     for operator in step.operators:
        #         print(type(operator))
        #         print(hasattr(operator, "pf"))

        self.add_qoi(
            name="pf",
            expr=sum([operator.pf*operator.measure for step in self.steps for operator in step.operators if hasattr(operator, "pf")]))



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin

import dolfin_mech as dmech
from .Material_Elastic import ElasticMaterial

################################################################################

class KirchhoffElasticMaterial(ElasticMaterial):



    def __init__(self,
            kinematics,
            parameters):

        self.kinematics = kinematics

        self.lmbda = self.get_lambda_from_parameters(parameters)
        self.mu    = self.get_mu_from_parameters(parameters)

        self.kinematics.E = dolfin.variable(self.kinematics.E)

        self.Psi = (self.lmbda/2) * dolfin.tr(self.kinematics.E)**2 + self.mu * dolfin.inner(self.kinematics.E, self.kinematics.E)

        self.Sigma = dolfin.diff(self.Psi, self.kinematics.E)
        # self.Sigma = self.lmbda * dolfin.tr(self.kinematics.E) * self.kinematics.I + 2 * self.mu * self.kinematics.E

        if (self.kinematics.dim == 2):
            self.Sigma_ZZ = self.lmbda * dolfin.tr(self.kinematics.E)

        self.P = self.kinematics.F * self.Sigma
        
        self.sigma = self.P * self.kinematics.F.T / self.kinematics.J



    # def get_free_energy(self,
    #         U=None,
    #         C=None,
    #         E=None):

    #     E = self.get_E_from_U_C_or_E(U, C, E)

    #     Psi = (self.lmbda/2) * dolfin.tr(E)**2 + self.mu * dolfin.inner(E, E)
    #     Sigma = dolfin.diff(Psi, E)

    #     # assert (E.ufl_shape[0] == E.ufl_shape[1])
    #     # dim = E.ufl_shape[0]
    #     # I = dolfin.Identity(dim)
    #     # Sigma = self.lmbda * dolfin.tr(E) * I + 2 * self.mu * E

    #     return Psi, Sigma

################################################################################

class KirchhoffBulkElasticMaterial(ElasticMaterial):



    def __init__(self,
            kinematics,
            parameters):

        self.kinematics = kinematics

        # self.K = self.get_K_from_parameters(parameters)
        self.lmbda, self.mu = self.get_lambda_and_mu_from_parameters(parameters)
        self.K = (self.kinematics.dim*self.lmbda + 2*self.mu)/self.kinematics.dim

        self.Psi   = (self.kinematics.dim*self.K/2) * dolfin.tr(self.kinematics.E_sph)**2
        self.Sigma =  self.kinematics.dim*self.K    *           self.kinematics.E_sph

        if (self.kinematics.dim == 2):
            self.Sigma_ZZ = self.K * dolfin.tr(self.kinematics.E)

        # self.P = dolfin.diff(self.Psi, self.kinematics.F) # MG20220426: Cannot do that for micromechanics problems
        self.P     = self.kinematics.F * self.Sigma

        self.sigma = self.P * self.kinematics.F.T / self.kinematics.J



    # def get_free_energy(self,
    #         U=None,
    #         C=None,
    #         E=None,
    #         E_sph=None):

    #     E_sph = self.get_E_sph_from_U_C_E_or_E_sph(
    #         U, C, E, E_sph)
    #     assert (E_sph.ufl_shape[0] == E_sph.ufl_shape[1])
    #     dim = E_sph.ufl_shape[0]

    #     Psi   = (dim*self.K/2) * dolfin.tr(E_sph)**2
    #     Sigma =  dim*self.K    *           E_sph

    #     return Psi, Sigma



    # def get_PK2_stress(self,
    #         U=None,
    #         C=None,
    #         E=None,
    #         E_sph=None):

    #     E_sph = self.get_E_sph_from_U_C_E_or_E_sph(
    #         U, C, E, E_sph)
    #     assert (E_sph.ufl_shape[0] == E_sph.ufl_shape[1])
    #     dim = E_sph.ufl_shape[0]

    #     Sigma = dim * self.K * E_sph

    #     return Sigma

################################################################################

class KirchhoffDevElasticMaterial(ElasticMaterial):



    def __init__(self,
            kinematics,
            parameters):

        self.kinematics = kinematics

        self.G = self.get_G_from_parameters(parameters)

        self.Psi   =   self.G * dolfin.inner(self.kinematics.E_dev, self.kinematics.E_dev)
        self.Sigma = 2*self.G *              self.kinematics.E_dev

        if (self.kinematics.dim == 2):
            self.Sigma_ZZ = -2*self.G/3 * dolfin.tr(self.kinematics.E)

        # self.P     = dolfin.diff(self.Psi, self.kinematics.F)
        self.P     = self.kinematics.F * self.Sigma

        self.sigma = self.P * self.kinematics.F.T / self.kinematics.J



    # def get_free_energy(self,
    #         U=None,
    #         C=None,
    #         E=None,
    #         E_dev=None):

    #     E_dev = self.get_E_dev_from_U_C_E_or_E_dev(
    #         U, C, E, E_dev)
        
    #     Psi   =   self.G * dolfin.inner(E_dev, E_dev)
    #     Sigma = 2*self.G *              E_dev

    #     return Psi, Sigma



    # def get_PK2_stress(self,
    #         U=None,
    #         C=None,
    #         E=None,
    #         E_dev=None):

    #     E_dev = self.get_E_dev_from_U_C_E_or_E_dev(
    #         U, C, E, E_dev)
        
    #     Sigma = 2*self.G * E_dev

    #     return Sigma



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin
import matplotlib.pyplot as mpl
import pandas
import numpy

import dolfin_mech as dmech

################################################################################

def run_RivlinCube_PoroHyperelasticity(
        dim=3,
        inverse=0,
        cube_params={},
        porosity_params={},
        move_params={},
        mat_params={},
        step_params={},
        load_params={},
        inertia_params={"applied": False},
        res_basename="run_RivlinCube_PoroHyperelasticity",
        plot_curves=False,
        get_results=0,
        verbose=0):

    ################################################################### Mesh ###

    if ("path_and_mesh_name" in cube_params):
        mesh = dolfin.Mesh()
        mesh_name = str(cube_params["path_and_mesh_name"])
        dolfin.XDMFFile(mesh_name).read(mesh)
        if ("refine" in cube_params):
            mesh=dolfin.refine(mesh)
        boundaries_mf = dolfin.MeshFunction("size_t", mesh, mesh.topology().dim()-1) # MG20180418: size_t looks like unisgned int, but more robust wrt architecture and os
        boundaries_mf.set_all(0)
    else:
        if   (dim==2):
            mesh, boundaries_mf, xmin_id, xmax_id, ymin_id, ymax_id = dmech.run_RivlinCube_Mesh(dim=dim, params=cube_params)
        elif (dim==3):
            mesh, boundaries_mf, xmin_id, xmax_id, ymin_id, ymax_id, zmin_id, zmax_id = dmech.run_RivlinCube_Mesh(dim=dim, params=cube_params)

    if move_params.get("move", False) == True :
        Umove = move_params.get("U")
        dolfin.ALE.move(mesh, Umove)

    ################################################################ Porosity ###

    porosity_type = porosity_params.get("type", "constant")
    porosity_val  = porosity_params.get("val", 0.5)

    if (porosity_type == "constant"):
        porosity_fun = None
    elif (porosity_type.startswith("mesh_function")):
        if (porosity_type == "mesh_function_constant"):
            porosity_mf = dolfin.MeshFunction(
                value_type="double",
                mesh=mesh,
                dim=dim,
                value=porosity_val)
        elif (porosity_type == "mesh_function_xml"):
            porosity_filename = res_basename+"-poro.xml"
            n_cells = len(mesh.cells())
            with open(porosity_filename, "w") as file:
                file.write('<?xml version="1.0" encoding="UTF-8"?>\n')
                file.write('<dolfin xmlns:dolfin="http://fenicsproject.org">\n')
                file.write('  <mesh_function type="double" dim="'+str(dim)+'" size="'+str(n_cells)+'">\n')
                for k_cell in range(n_cells):
                    file.write('    <entity index="'+str(k_cell)+'" value="'+str(porosity_val)+'"/>\n')
                file.write('  </mesh_function>\n')
                file.write('</dolfin>\n')
                file.close()
            porosity_mf = dolfin.MeshFunction(
                "double",
                mesh,
                porosity_filename)
        elif (porosity_type == "mesh_function_random_xml"):
            # print("mesh_function xml")
            porosity_filename = res_basename+"-poro.xml"
            n_cells = len(mesh.cells())
            with open(porosity_filename, "w") as file:
                file.write('<?xml version="1.0" encoding="UTF-8"?>\n')
                file.write('<dolfin xmlns:dolfin="http://fenicsproject.org">\n')
                file.write('  <mesh_function type="double" dim="'+str(dim)+'" size="'+str(n_cells)+'">\n')
                for k_cell in range(n_cells):
                    value = numpy.random.uniform(low=0.4, high=0.6)
                            # positive_value = True
                    file.write('    <entity index="'+str(k_cell)+'" value="'+str(value)+'"/>\n')
                file.write('  </mesh_function>\n')
                file.write('</dolfin>\n')
                file.close()
            porosity_mf = dolfin.MeshFunction(
                "double",
                mesh,
                porosity_filename)
        porosity_expr = dolfin.CompiledExpression(getattr(dolfin.compile_cpp_code(dmech.get_ExprMeshFunction_cpp_pybind()), "MeshExpr")(), mf=porosity_mf, degree=0)
        porosity_fs = dolfin.FunctionSpace(mesh, 'DG', 0)
        porosity_fun = dolfin.interpolate(porosity_expr, porosity_fs)
        porosity_val = None
    elif (porosity_type.startswith("function")):
        porosity_fs = dolfin.FunctionSpace(mesh, 'DG', 0)
        if (porosity_type == "function_constant"):
            porosity_fun = dolfin.Function(porosity_fs)
            porosity_fun.vector()[:] = porosity_val
        elif (porosity_type == "function_xml"):
            porosity_filename = res_basename+"-poro.xml"
            n_cells = len(mesh.cells())
            with open(porosity_filename, "w") as file:
                file.write('<?xml version="1.0" encoding="UTF-8"?>\n')
                file.write('<dolfin xmlns:dolfin="http://fenicsproject.org">\n')
                file.write('  <function_data size="'+str(n_cells)+'">\n')
                for k_cell in range(n_cells):
                    file.write('    <dof index="'+str(k_cell)+'" value="'+str(porosity_val)+'" cell_index="'+str(k_cell)+'" cell_dof_index="0"/>\n')
                file.write('  </function_data>\n')
                file.write('</dolfin>\n')
                file.close()
            porosity_fun = dolfin.Function(
                porosity_fs,
                porosity_filename)
        elif (porosity_type == "function_xml_from_array"):
            # print("function xml")
            porosity_filename = res_basename+"-poro.xml"
            # print("porosity_filename=", porosity_filename)
            n_cells = len(mesh.cells())
            with open(porosity_filename, "w") as file:
                file.write('<?xml version="1.0" encoding="UTF-8"?>\n')
                file.write('<dolfin xmlns:dolfin="http://fenicsproject.org">\n')
                file.write('  <function_data size="'+str(n_cells)+'">\n')
                for k_cell in range(n_cells):
                    # print("kcell=", k_cell)
                    # print("porosity for kcell", porosity_val[k_cell])
                    file.write('    <dof index="'+str(k_cell)+'" value="'+str(porosity_val[k_cell])+'" cell_index="'+str(k_cell)+'" cell_dof_index="0"/>\n')
                file.write('  </function_data>\n')
                file.write('</dolfin>\n')
                file.close()
            porosity_fun = dolfin.Function(
                porosity_fs,
                porosity_filename)
        porosity_val = None

    ################################################################ Problem ###
    load_type = load_params.get("type", "internal")

    if load_type=="p_boundary_condition0":
        gradient_operators="inverse"
    elif load_type=="p_boundary_condition":
        gradient_operators="direct"
    else:
        gradient_operators=None

    if (inverse):
        problem = dmech.InversePoroHyperelasticityProblem(
            mesh=mesh,
            define_facet_normals=1,
            boundaries_mf=boundaries_mf,
            displacement_degree=1,
            porosity_init_val=porosity_val,
            porosity_init_fun=porosity_fun,
            skel_behavior=mat_params,
            bulk_behavior=mat_params,
            pore_behavior=mat_params,
            gradient_operators=gradient_operators)
    else:
        problem = dmech.PoroHyperelasticityProblem(
            mesh=mesh,
            define_facet_normals=1,
            boundaries_mf=boundaries_mf,
            displacement_degree=1,
            porosity_init_val=porosity_val,
            porosity_init_fun=porosity_fun,
            skel_behavior=mat_params,
            bulk_behavior=mat_params,
            pore_behavior=mat_params,
            gradient_operators=gradient_operators)

    ########################################## Boundary conditions & Loading ###
    n_steps = step_params.get("n_steps", 1)
    Deltat_lst = step_params.get("Deltat_lst", [step_params.get("Deltat", 1.)/n_steps]*n_steps)
    dt_ini_lst = step_params.get("dt_ini_lst", [step_params.get("dt_ini", 1.)/n_steps]*n_steps)
    dt_min_lst = step_params.get("dt_min_lst", [step_params.get("dt_min", 1.)/n_steps]*n_steps)

    load_type = load_params.get("type", "internal")

    if   (load_type == "internal"):
        pf_lst = load_params.get("pf_lst", [(k_step+1)*load_params.get("pf", +0.5)/n_steps for k_step in range(n_steps)])
    elif (load_type in ("external0", "external")):
        P_lst = load_params.get("P_lst", [(k_step+1)*load_params.get("P", -0.5)/n_steps for k_step in range(n_steps)])
    elif (load_type in ("p_boundary_condition0", "p_boundary_condition")):
        f_lst = load_params.get("f_lst", [(k_step+1)*load_params.get("f", 1e4)/n_steps for k_step in range(n_steps)])
        P0_lst = load_params.get("P0_lst", [(k_step+1)*load_params.get("P0", -0.50)/n_steps for k_step in range(n_steps)])
    
    for k_step in range(n_steps):

        Deltat = Deltat_lst[k_step]
        dt_ini = dt_ini_lst[k_step]
        dt_min = dt_min_lst[k_step]

        k_step = problem.add_step(
            Deltat=Deltat,
            dt_ini=dt_ini,
            dt_min=dt_min)

        if (load_type == "internal"):
            pf = pf_lst[k_step]
            pf_old = pf_lst[k_step-1] if (k_step > 0) else 0.
            problem.add_pf_operator(
                measure=problem.dV,
                pf_ini=pf_old, pf_fin=pf,
                k_step=k_step)
        elif (load_type == "external"):
            problem.add_pf_operator(
                measure=problem.dV,
                pf_ini=0., pf_fin=0.,
                k_step=k_step)
            P = P_lst[k_step]
            P_old = P_lst[k_step-1] if (k_step > 0) else 0.
            problem.add_surface_pressure_loading_operator(
                measure=problem.dS(xmax_id),
                P_ini=P_old, P_fin=P,
                k_step=k_step)
            problem.add_surface_pressure_loading_operator(
                measure=problem.dS(ymax_id),
                P_ini=P_old, P_fin=P,
                k_step=k_step)
            if (dim==3): problem.add_surface_pressure_loading_operator(
                measure=problem.dS(zmax_id),
                P_ini=P_old, P_fin=P,
                k_step=k_step)
        elif (load_type == "external0"):
            problem.add_pf_operator(
                measure=problem.dV,
                pf_ini=0., pf_fin=0.,
                k_step=k_step)
            P = P_lst[k_step]
            P_old = P_lst[k_step-1] if (k_step > 0) else 0.
            problem.add_surface_pressure0_loading_operator(
                measure=problem.dS(xmax_id),
                P_ini=P_old, P_fin=P,
                k_step=k_step)
            problem.add_surface_pressure0_loading_operator(
                measure=problem.dS(ymax_id),
                P_ini=P_old, P_fin=P,
                k_step=k_step)
            if (dim==3): problem.add_surface_pressure0_loading_operator(
                measure=problem.dS(zmax_id),
                P_ini=P_old, P_fin=P,
                k_step=k_step)
        elif (load_type == "p_boundary_condition0"):
            problem.add_pf_operator(
                measure=problem.dV,
                pf_ini=0.,
                pf_fin=0.,
                k_step=k_step)
            f = f_lst[k_step]
            f_old = f_lst[k_step-1] if (k_step > 0) else 0.
            P0 = P0_lst[k_step]
            P0_old = P0_lst[k_step-1] if (k_step > 0) else 0.
            rho_solid = mat_params.get("parameters").get("rho_solid", 1e-6)
            problem.add_pressure_balancing_gravity0_loading_operator(
                dV=problem.dV,
                dS=problem.dS,
                f_ini=[0.]*(dim-1)+[f_old],
                # f_fin=[f, 0., 0.],
                f_fin=[0., 0., f],
                rho_solid=rho_solid,
                phis=problem.phis,
                P0_ini=P0_old,
                P0_fin=P0,
                breathing_constant=load_params.get("H", 0.),
                k_step=k_step)
        elif (load_type == "p_boundary_condition"):
            problem.add_pf_operator(
                measure=problem.dV,
                pf_ini=0.,
                pf_fin=0.,
                k_step=k_step)
            f = f_lst[k_step]
            f_old = f_lst[k_step-1] if (k_step > 0) else 0.
            P0 = P0_lst[k_step]
            P0_old = P0_lst[k_step-1] if (k_step > 0) else 0.
            rho_solid = mat_params.get("parameters").get("rho_solid", 1e-6)
            problem.add_pressure_balancing_gravity_loading_operator(
                dV=problem.dV,
                dS=problem.dS,
                f_ini=[0.]*(dim-1)+[f_old],
                # f_fin=[f, 0., 0.],
                f_fin=[0., 0., f],
                rho_solid=rho_solid,
                P0_ini=P0_old,
                P0_fin=P0,
                breathing_constant=load_params.get("H", 0.),
                k_step=k_step)
        
        if not inertia_params["applied"]:
            problem.add_constraint(V=problem.get_displacement_function_space().sub(0), sub_domains=boundaries_mf, sub_domain_id=xmin_id, val=0.)
            problem.add_constraint(V=problem.get_displacement_function_space().sub(1), sub_domains=boundaries_mf, sub_domain_id=ymin_id, val=0.)
            if (dim==3):
                problem.add_constraint(V=problem.get_displacement_function_space().sub(2), sub_domains=boundaries_mf, sub_domain_id=zmin_id, val=0.)
        else:
            rho_val=inertia_params.get("rho_val", 1e-6)
            problem.add_inertia_operator(
            measure=problem.dV,
            rho_val=rho_val,
            k_step=k_step)

    ################################################# Quantities of Interest ###

    problem.add_deformed_volume_qoi()
    problem.add_global_strain_qois()
    problem.add_global_stress_qois()
    problem.add_global_porosity_qois()
    problem.add_global_fluid_pressure_qoi()

    ################################################################# Solver ###

    solver = dmech.NonlinearSolver(
        problem=problem,
        parameters={
            "sol_tol":[1e-6]*len(problem.subsols),
            "n_iter_max":32},
        relax_type="constant",
        write_iter=0)

    integrator = dmech.TimeIntegrator(
        problem=problem,
        solver=solver,
        parameters={
            "n_iter_for_accel":4,
            "n_iter_for_decel":16,
            "accel_coeff":2,
            "decel_coeff":2},
        print_out=res_basename*verbose,
        print_sta=res_basename*verbose,
        write_qois=res_basename+"-qois",
        write_qois_limited_precision=1,
        write_sol=res_basename*verbose)

    success = integrator.integrate()
    assert (success),\
        "Integration failed. Aborting."

    integrator.close()

    ################################################################## Plots ###

    if (plot_curves):
        qois_data = pandas.read_csv(
            res_basename+"-qois.dat",
            delim_whitespace=True,
            comment="#",
            names=open(res_basename+"-qois.dat").readline()[1:].split())

        qois_fig, qois_axes = mpl.subplots()
        all_strains = ["E_XX", "E_YY"]
        if (dim == 3): all_strains += ["E_ZZ"]
        all_strains += ["E_XY"]
        if (dim == 3): all_strains += ["E_YZ", "E_ZX"]
        qois_data.plot(x="t", y=all_strains, ax=qois_axes, ylabel="Green-Lagrange strain")
        qois_fig.savefig(res_basename+"-strains-vs-time.pdf")

        for comp in ["skel", "bulk", "tot"]:
            qois_fig, qois_axes = mpl.subplots()
            all_stresses = ["s_"+comp+"_XX", "s_"+comp+"_YY"]
            if (dim == 3): all_stresses += ["s_"+comp+"_ZZ"]
            all_stresses += ["s_"+comp+"_XY"]
            if (dim == 3): all_stresses += ["s_"+comp+"_YZ", "s_"+comp+"_ZX"]
            qois_data.plot(x="t", y=all_stresses, ax=qois_axes, ylabel="Cauchy stress")
            qois_fig.savefig(res_basename+"-stresses-"+comp+"-vs-time.pdf")

        qois_fig, qois_axes = mpl.subplots()
        all_porosities = []
        if (inverse):
            all_porosities += ["phis0", "phif0", "Phis0", "Phif0"]
        else:
            all_porosities += ["Phis", "Phif", "phis", "phif"]
        qois_data.plot(x="t", y=all_porosities, ax=qois_axes, ylim=[0,1], ylabel="porosity")
        qois_fig.savefig(res_basename+"-porosities-vs-time.pdf")

        qois_fig, qois_axes = mpl.subplots()
        qois_data.plot(x="pf", y=all_porosities, ax=qois_axes, ylim=[0,1], ylabel="porosity")
        qois_fig.savefig(res_basename+"-porosities-vs-pressure.pdf")
    
    if get_results:
        if inverse:
            phi = problem.get_foi(name="Phis0").func.vector().get_local()
        else:
            phi = problem.get_foi(name="phis").func.vector().get_local()
        return (problem.get_displacement_subsol().func, phi, dolfin.Measure("dx", domain=mesh))



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
###                                                                          ###
### And Mahdi Manoochehrtayebi, 2020-2024                                    ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin

import dolfin_mech as dmech
from .Operator import Operator

################################################################################

class MacroscopicStressComponentPenaltyOperator(Operator):

    def __init__(self,
            sigma_bar,
            sigma_bar_test,
            sol,
            sol_test,
            material,
            comp_i, comp_j,
            measure,
            comp_val=None, comp_ini=None, comp_fin=None,
            pen_val=None, pen_ini=None, pen_fin=None):

        self.material = material
        self.measure  = measure

        self.tv_comp = dmech.TimeVaryingConstant(
            val=comp_val, val_ini=comp_ini, val_fin=comp_fin)
        comp = self.tv_comp.val

        self.tv_pen = dmech.TimeVaryingConstant(
            val=pen_val, val_ini=pen_ini, val_fin=pen_fin)
        pen = self.tv_pen.val

        Pi = (pen/2) * (self.material.sigma[comp_i,comp_j] - comp)**2 * self.measure # MG20220426: Need to compute <sigma> properly, including fluid pressure
        # self.res_form = dolfin.derivative(Pi, sigma_bar[comp_i,comp_j], sigma_bar_test[comp_i,comp_j]) # MG20230106: This does not work…
        self.res_form = dolfin.derivative(Pi, sol, sol_test) # MG20230106: This works…

        # Pi = (pen/2) * (sigma_bar[comp_i,comp_j] - comp)**2 * self.measure # MG20230106: This does not work…
        # self.res_form = dolfin.derivative(Pi, sigma_bar[comp_i,comp_j], sigma_bar_test[comp_i,comp_j])
        # self.res_form = dolfin.derivative(Pi, sol, sol_test)



    def set_value_at_t_step(self,
            t_step):

        self.tv_comp.set_value_at_t_step(t_step)
        self.tv_pen.set_value_at_t_step(t_step)



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
###                                                                          ###
### And Mahdi Manoochehrtayebi, 2020-2024                                    ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin 

import dolfin_mech as dmech
from .Material_Elastic import ElasticMaterial

################################################################################

class OgdenCiarletGeymonatNeoHookeanElasticMaterial(ElasticMaterial):



    def __init__(self,
            kinematics,
            parameters,
            decoup=False):

        self.kinematics = kinematics

        self.bulk = dmech.OgdenCiarletGeymonatElasticMaterial(kinematics, parameters, decoup)
        self.dev  = dmech.NeoHookeanElasticMaterial(kinematics, parameters, decoup)

        self.Psi   = self.bulk.Psi   + self.dev.Psi
        self.Sigma = self.bulk.Sigma + self.dev.Sigma
        if (self.kinematics.dim == 2):
            self.Sigma_ZZ = self.bulk.Sigma_ZZ + self.dev.Sigma_ZZ
            self.p_hydro = -(dolfin.tr(self.Sigma.T*self.kinematics.C)+ self.Sigma_ZZ)/3/self.kinematics.J
        elif (self.kinematics.dim == 3):
            self.p_hydro = -(dolfin.tr(self.Sigma.T*self.kinematics.C))/3/self.kinematics.J
        self.P     = self.bulk.P     + self.dev.P
        self.sigma = self.bulk.sigma + self.dev.sigma
        self.Sigma_dev = self.Sigma + self.p_hydro * self.kinematics.J * self.kinematics.C_inv
        self.Sigma_VM = dolfin.sqrt(1.5 *dolfin.tr(self.Sigma_dev.T*self.Sigma_dev))



    # def get_free_energy(self, *args, **kwargs):

    #     Psi_bulk, Sigma_bulk = self.bulk.get_free_energy(*args, **kwargs)
    #     Psi_dev , Sigma_dev  = self.dev.get_free_energy(*args, **kwargs)

    #     Psi   = Psi_bulk   + Psi_dev
    #     Sigma = Sigma_bulk + Sigma_dev

    #     return Psi, Sigma



    # def get_PK2_stress(self, *args, **kwargs):

    #     Sigma_bulk = self.bulk.get_PK2_stress(*args, **kwargs)
    #     Sigma_dev  = self.dev.get_PK2_stress(*args, **kwargs)

    #     Sigma = Sigma_bulk + Sigma_dev

    #     return Sigma



    # def get_PK1_stress(self, *args, **kwargs):

    #     P_bulk = self.bulk.get_PK1_stress(*args, **kwargs)
    #     P_dev  = self.dev.get_PK1_stress(*args, **kwargs)

    #     P = P_bulk + P_dev

    #     return P



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin

import dolfin_mech as dmech

################################################################################

class Constraint():



    def __init__(self,
            V,
            sub_domain=None,
            sub_domains=None,
            sub_domain_id=None,
            val=None, val_ini=None, val_fin=None,
            method=None): # topological, geometric, pointwise

        if (val is not None) and (val_ini is None) and (val_fin is None):
            self.tv_val = dmech.TimeVaryingConstant(
                val_ini=val, val_fin=val)
        elif (val is None) and (val_ini is not None) and (val_fin is not None):
            self.tv_val = dmech.TimeVaryingConstant(
                val_ini=val_ini, val_fin=val_fin)

        if (sub_domain is not None) and (sub_domains is None) and (sub_domain_id is None):
            if (method is None):
                self.bc = dolfin.DirichletBC(
                    V,
                    self.tv_val.val,
                    sub_domain)
            else:
                self.bc = dolfin.DirichletBC(
                    V,
                    self.tv_val.val,
                    sub_domain,
                    method)
        elif (sub_domain is None) and (sub_domains is not None) and (sub_domain_id is not None):
            if (method is None):
                self.bc = dolfin.DirichletBC(
                    V,
                    self.tv_val.val,
                    sub_domains,
                    sub_domain_id)
            else:
                self.bc = dolfin.DirichletBC(
                    V,
                    self.tv_val.val,
                    sub_domains,
                    sub_domain_id,
                    method)



    def set_value_at_t_step(self,
            t_step):

        self.tv_val.set_dvalue_at_t_step(t_step)



    def restore_old_value(self):

        self.tv_val.restore_old_value()



    def homogenize(self):

        self.tv_val.homogenize()
        # self.bc.homogenize() # MG20180508: seems to be changing the constant…



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import decimal
import dolfin
import glob
import math
import numpy
import os
import petsc4py
import petsc4py.PETSc
import sys
import time

import myPythonLibrary as mypy

import dolfin_mech as dmech

################################################################################

class NonlinearSolver():



    def __init__(self,
            problem,
            parameters,
            relax_type="constant",
            relax_parameters={},
            print_out=True,
            write_iter=False):

        self.problem = problem

        self.default_linear_solver_type = "petsc"
        #self.default_linear_solver_type = "dolfin"

        self.linear_solver_type = parameters.get("linear_solver_type", self.default_linear_solver_type)

        if (self.linear_solver_type == "petsc"):

            self.res_vec = dolfin.PETScVector()
            self.jac_mat = dolfin.PETScMatrix()

            self.linear_solver = dolfin.PETScKrylovSolver()

            self.default_linear_solver_name = "mumps"

            self.linear_solver_name = parameters.get("linear_solver_name", self.default_linear_solver_name)

            if (self.linear_solver_name == "mumps"):
                if (int(dolfin.__version__.split('.')[0]) >= 2018):
                    options = petsc4py.PETSc.Options()
                    options["ksp_type"] = "preonly"
                    options["pc_type"] = "lu"
                    options["pc_factor_mat_solver_type"] = "mumps"
                    options["mat_mumps_icntl_33"] = 0
                else:
                    options = dolfin.PETScOptions()
                    options.set("ksp_type", "preonly")
                    options.set("pc_type", "lu")
                    options.set("pc_factor_mat_solver_package", "mumps")
                    options.set("mat_mumps_icntl_33", 0)

            self.linear_solver.ksp().setFromOptions()
            self.linear_solver.ksp().setOperators(A=self.jac_mat.mat())

        elif (self.linear_solver_type == "dolfin"):

            self.res_vec = dolfin.Vector()
            self.jac_mat = dolfin.Matrix()

            # self.default_linear_solver_name = "default"
            self.default_linear_solver_name = "mumps"
            # self.default_linear_solver_name = "petsc"
            # self.default_linear_solver_name = "superlu"
            # self.default_linear_solver_name = "umfpack"

            self.linear_solver_name = parameters.get("linear_solver_name", self.default_linear_solver_name)

            self.linear_solver = dolfin.LUSolver(
                self.jac_mat,
                self.linear_solver_name)
            # self.linear_solver.parameters['report']               = bool(0)
            # self.linear_solver.parameters['reuse_factorization']  = bool(0)
            # self.linear_solver.parameters['same_nonzero_pattern'] = bool(1)
            # self.linear_solver.parameters['symmetric']            = bool(1)
            # self.linear_solver.parameters['verbose']              = bool(1)

        if (relax_type == "constant"):
            self.compute_relax = self.compute_relax_constant
            self.relax_val = relax_parameters.get("relax", 1.)
        elif (relax_type == "aitken"):
            self.compute_relax = self.compute_relax_aitken
        elif (relax_type == "gss"):
            self.compute_relax = self.compute_relax_gss
            self.relax_n_iter_max = relax_parameters.get("relax_n_iter_max", 9)
        elif (relax_type == "backtracking"):
            self.compute_relax = self.compute_relax_backtracking
            self.relax_backtracking_factor = parameters["relax_backtracking_factor"] if ("relax_backtracking_factor" in parameters) and (parameters["relax_backtracking_factor"] is not None) else 2.
            self.relax_n_iter_max          = parameters["relax_n_iter_max"]          if ("relax_n_iter_max"          in parameters) and (parameters["relax_n_iter_max"]          is not None) else 8

        self.sol_tol    = parameters.get("sol_tol"   , [1e-6]*len(self.problem.subsols))
        self.n_iter_max = parameters.get("n_iter_max", 32)

        if (type(print_out) is str):
            if (print_out=="stdout"):
                self.printer_filename = None
            elif (print_out=="argv"):
                self.printer_filename = sys.argv[0][:-3]+".out"
            else:
                self.printer_filename = print_out+".out"
        else:
            self.printer_filename = None
        self.printer = mypy.Printer(
            filename=self.printer_filename,
            silent=not(print_out))

        self.write_iter = bool(write_iter)
        if (self.write_iter):
            for filename in glob.glob(sys.argv[0][:-3]+"-sol-k_step=*-k_t=*.*"):
                os.remove(filename)

            self.functions_to_write = []
            self.functions_to_write += self.problem.get_subsols_func_lst()
            self.functions_to_write += self.problem.get_subsols_func_old_lst()
            self.functions_to_write += self.problem.get_fois_func_lst()



    def solve(self,
            k_step=None,
            k_t=None,
            dt=None,
            t=None):

        # write
        if (self.write_iter):
            xdmf_file_iter = dmech.XDMFFile(
                filename=sys.argv[0][:-3]+"-sol-k_step="+str(k_step)+"-k_t="+str(k_t)+".xdmf",
                functions=self.functions_to_write)
            self.problem.update_fois()
            xdmf_file_iter.write(0.)

        self.k_iter = 0
        self.success = False
        self.printer.inc()
        while (True):
            self.k_iter += 1
            self.printer.print_var("k_iter",self.k_iter,-1)

            # linear problem
            linear_success = self.linear_solve(
                k_step=k_step,
                k_t=k_t)
            if not (linear_success):
                break
            self.compute_dsol_norm()

            # constraints update
            if (self.k_iter == 1):
                for constraint in self.constraints:
                    constraint.homogenize()

            # solution update
            self.compute_relax()
            self.update_sol()
            self.compute_sol_norm()

            # internal variables update
            for inelastic_behavior in self.problem.inelastic_behaviors_internal:
                inelastic_behavior.update_internal_variables_after_solve(
                    dt, t)

            # write
            if (self.write_iter):
                self.problem.update_fois()
                xdmf_file_iter.write(self.k_iter)

            # error
            self.compute_sol_err()

            # exit test
            self.exit_test()

            if (self.success):
                self.printer.print_str("Nonlinear solver converged…")
                break

            if (self.k_iter == self.n_iter_max):
                self.printer.print_str("Warning! Nonlinear solver failed to converge!")
                break

        self.printer.dec()

        # write
        if (self.write_iter):
            xdmf_file_iter.close()

        return self.success, self.k_iter



    def linear_solve(self,
            k_step=None,
            k_t=None):

        assemble_linear_system = self.assemble_linear_system()   

        if assemble_linear_system==False:
            return False 
        
        # eigen problem
        if (k_step == 1) and (k_t == 1) and (self.k_iter == 1) and (0):
            self.eigen_solve()

        # linear system: solve
        try:
            self.printer.print_str("Solve…",newline=False)
            timer = time.time()
            self.linear_solver.solve(
                self.problem.dsol_func.vector(),
                self.res_vec)
            timer = time.time() - timer
            self.printer.print_str(" "+str(timer)+" s",tab=False)
            #self.printer.print_var("dsol_func",self.problem.dsol_func.vector().get_local())
        except:
            self.printer.print_str("Warning! Linear solver failed!",tab=False)
            return False

        if not (numpy.isfinite(self.problem.dsol_func.vector()).all()):
            # self.problem.dsol_func.vector().zero()

            self.printer.print_str("Warning! Solution increment is NaN!")
            return False

        if (len(self.problem.subsols) > 1):
            dolfin.assign(
                self.problem.get_subsols_dfunc_lst(),
                self.problem.dsol_func)
            # for subsol_name,subsol in self.problem.subsols.items():
            #     self.printer.print_var("d"+subsol_name+"_func",subsol.dfunc.vector().get_local())

        if (0):
            rinfo12 = self.linear_solver.ksp().getPC().getFactorMatrix().getMumpsRinfog(12)
            #self.printer.print_sci("rinfo12",rinfo12)
            rinfo12 = decimal.Decimal(rinfo12)
            #self.printer.print_sci("rinfo12",rinfo12)
            infog34 = self.linear_solver.ksp().getPC().getFactorMatrix().getMumpsInfog(34)
            #self.printer.print_sci("infog34",infog34)
            infog34 = decimal.Decimal(infog34)
            #self.printer.print_sci("infog34",infog34)
            self.jac_det = rinfo12*(decimal.Decimal(2.)**infog34)
            self.printer.print_sci("jac_det",self.jac_det)

        return True

    def assemble_linear_system(self):

        # res_old
        if (self.k_iter > 1):
            if (hasattr(self, "res_old_vec")):
                self.res_old_vec[:] = self.res_vec[:]
            else:
                self.res_old_vec = self.res_vec.copy()
            self.res_old_norm = self.res_norm

        # linear system: Assembly
        if any([(operator.measure.integral_type() == "vertex") for operator in self.problem.operators]): # MG20190513: Cannot use point integral within assemble_system
            self.printer.print_str("Assembly (without vertex integrals)…",newline=False)
            timer = time.time()
            dolfin.assemble_system(
                self.problem.jac_form,
               -self.problem.res_form,
                bcs=[constraint.bc for constraint in self.constraints],
                A_tensor=self.jac_mat,
                b_tensor=self.res_vec,
                add_values=False,
                finalize_tensor=False,
                form_compiler_parameters=self.problem.form_compiler_parameters)
            timer = time.time() - timer
            self.printer.print_str(" "+str(timer)+" s",tab=False)
            # self.printer.print_var("res_vec",self.res_vec.get_local())
            # self.printer.print_var("jac_mat",self.jac_mat.array())

            for operator in self.problem.operators:
                if (operator.measure.integral_type() == "vertex"):
                    self.printer.print_str("Assembly (vertex integrals)…",newline=False)
                    timer = time.time()
                    dolfin.assemble( # MG20190513: However, vertex integrals only work if solution only has dofs on vertices…
                       -operator.res_form,
                        tensor=self.res_vec,
                        add_values=True,
                        finalize_tensor=True,
                        form_compiler_parameters=self.problem.form_compiler_parameters)
                    operator.jac_form = dolfin.derivative(
                        operator.res_form,
                        self.problem.sol_func,
                        self.problem.dsol_tria)
                    dolfin.assemble(
                        operator.jac_form,
                        tensor=self.jac_mat,
                        add_values=True,
                        finalize_tensor=True,
                        form_compiler_parameters=self.problem.form_compiler_parameters)
                    timer = time.time() - timer
                    self.printer.print_str(" "+str(timer)+" s",tab=False)
                    # self.printer.print_var("res_vec",self.res_vec.get_local())
                    # self.printer.print_var("jac_mat",self.jac_mat.array())
        else:
            self.printer.print_str("Assembly…",newline=False)
            timer = time.time()
            dolfin.assemble_system(
                self.problem.jac_form,
               -self.problem.res_form,
                bcs=[constraint.bc for constraint in self.constraints],
                A_tensor=self.jac_mat,
                b_tensor=self.res_vec,
                add_values=False,
                finalize_tensor=True,
                form_compiler_parameters=self.problem.form_compiler_parameters)
            timer = time.time() - timer
            self.printer.print_str(" "+str(timer)+" s",tab=False)
            # self.printer.print_var("res_vec",self.res_vec.get_local())
            # self.printer.print_var("jac_mat",self.jac_mat.array())

        if not (numpy.isfinite(self.res_vec).all()):
            self.printer.print_str("Warning! Residual is NaN!")
            return False

        # res_norm
        self.res_norm = self.res_vec.norm("l2")
        self.printer.print_sci("res_norm",self.res_norm)

        if (self.res_norm > 1e9):
            self.printer.print_str("Warning! Residual is too large!")
            return False

        # res_err
        if (self.k_iter == 1):
            self.res_norm0 = self.res_norm
        else:
            self.res_err = dmech.compute_error(
                val=self.res_norm,
                ref=self.res_norm0)
            self.printer.print_sci("res_err",self.res_err)

            if (self.res_err > 1e3):
                self.printer.print_str("Warning! Residual is increasing too much!")
                return False

        # dres
        if (self.k_iter > 1):
            if (hasattr(self, "dres_vec")):
                self.dres_vec[:] = self.res_vec[:] - self.res_old_vec[:]
            else:
                self.dres_vec = self.res_vec - self.res_old_vec
            self.dres_norm = self.dres_vec.norm("l2")
            self.printer.print_sci("dres_norm",self.dres_norm)

        # res_err_rel
        if (self.k_iter > 1):
            self.res_err_rel = dmech.compute_error(
                val=self.dres_norm,
                ref=self.res_old_norm)
            self.printer.print_sci("res_err_rel",self.res_err_rel)


    def eigen_solve():

        jac_eigensolver = dolfin.SLEPcEigenSolver(
            dolfin.as_backend_type(self.jac_mat))

        # jac_eigensolver.parameters["problem_type"] = "non_hermitian"
        jac_eigensolver.parameters["problem_type"] = "hermitian"

        jac_eigensolver.parameters["solver"] = "krylov-schur"
        # jac_eigensolver.parameters["solver"] = "power"
        # jac_eigensolver.parameters["solver"] = "subspace"
        # jac_eigensolver.parameters["solver"] = "arnoldi"
        # jac_eigensolver.parameters["solver"] = "lanczos"

        # jac_eigensolver.parameters["tolerance"] = 1e-1
        # jac_eigensolver.parameters["maximum_iterations"] = 100

        jac_eigensolver.parameters["verbose"] = True

        mode_func = dolfin.Function(self.problem.sol_fs)

        n_modes = 10
        spectrums  = []
        # spectrums += ["largest"]
        spectrums += ["smallest"]

        for spectrum in spectrums:
            jac_eigensolver.parameters["spectrum"] = spectrum+" magnitude"

            self.printer.print_str("Eigenproblem solve…",newline=False)
            timer = time.time()
            jac_eigensolver.solve(n_modes)
            timer = time.time() - timer
            self.printer.print_str(" "+str(timer)+" s",tab=False,newline=False)

            n_converged = jac_eigensolver.get_number_converged()
            self.printer.print_str(" ("+str(n_converged)+" converged modes)",tab=False)

            xdmf_file_modes = dmech.XDMFFile(
                filename=sys.argv[0][:-3]+"-eigenmodes-"+spectrum+".xdmf",
                functions=[mode_func])
            for k_mode in range(n_converged):
                # print(k_mode+1)
                val_r, val_c, vec_r, vec_c = jac_eigensolver.get_eigenpair(k_mode)
                # print(val_r)
                mode_func.vector()[:] = vec_r[:]
                xdmf_file_modes.write(k_mode)
            xdmf_file_modes.close()



    def compute_dsol_norm(self):

        self.dsubsol_norm_lst = [subsol.dfunc.vector().norm("l2") for subsol in self.problem.subsols.values()]
        for (k_subsol,subsol) in enumerate(self.problem.subsols.values()):
            self.printer.print_sci("d"+subsol.name+"_norm",self.dsubsol_norm_lst[k_subsol])



    def compute_relax_constant(self):

        if (self.k_iter == 1):
            self.relax = 1. # MG20180505: Otherwise Dirichlet boundary conditions are not correctly enforced
        else:
            self.relax = self.relax_val
            self.printer.print_sci("relax",self.relax)



    def compute_relax_aitken(self):

        if (self.k_iter == 1):
            self.relax = 1. # MG20180505: Otherwise Dirichlet boundary conditions are not correctly enforced
        else:
            self.relax *= (-1.) * self.res_old_vec.inner(self.dres_vec) / self.dres_norm**2
        self.printer.print_sci("relax",self.relax)



    def compute_relax_gss(self):

        if (self.k_iter == 1):
            self.relax = 1. # MG20180505: Otherwise Dirichlet boundary conditions are not correctly enforced
        else:
            phi = (1+math.sqrt(5))/2
            a = (1-phi)/(2-phi)
            b = 1./(2-phi)
            need_update_c = True
            need_update_d = True
            cur = 0.
            relax_list = []
            relax_vals = []
            self.printer.inc()
            relax_k = 0
            while (True):
                self.printer.print_var("relax_k",relax_k)
                self.printer.print_sci("a",a)
                self.printer.print_sci("b",b)
                if (need_update_c):
                    c = b - (b - a) / phi
                    relax_list.append(c)
                    self.printer.print_sci("c",c)
                    self.problem.sol_func.vector().axpy(
                        c-cur,
                        self.problem.dsol_func.vector())
                    if (len(self.problem.subsols) > 1):
                        dolfin.assign(
                            self.problem.get_subsols_func_lst(),
                            self.problem.sol_func)
                    cur = c
                    relax_fc  = dolfin.assemble(
                        self.problem.Pi_expr,
                        form_compiler_parameters=self.problem.form_compiler_parameters)
                    #self.printer.print_sci("relax_fc",relax_fc)
                    if (numpy.isnan(relax_fc)):
                        relax_fc = float("+inf")
                        #self.printer.print_sci("relax_fc",relax_fc)
                    self.printer.print_sci("relax_fc",relax_fc)
                    relax_vals.append(relax_fc)
                    #self.printer.print_var("relax_list",relax_list)
                    #self.printer.print_var("relax_vals",relax_vals)
                if (need_update_d):
                    d = a + (b - a) / phi
                    relax_list.append(d)
                    self.printer.print_sci("d",d)
                    self.problem.sol_func.vector().axpy(
                        d-cur,
                        self.problem.dsol_func.vector())
                    if (len(self.problem.subsols) > 1):
                        dolfin.assign(
                            self.problem.get_subsols_func_lst(),
                            self.problem.sol_func)
                    cur = d
                    relax_fd  = dolfin.assemble(
                        self.problem.Pi_expr,
                        form_compiler_parameters=self.problem.form_compiler_parameters)
                    if (numpy.isnan(relax_fd)):
                        relax_fd = float("+inf")
                        #self.printer.print_sci("relax_fd",relax_fd)
                    self.printer.print_sci("relax_fd",relax_fd)
                    relax_vals.append(relax_fd)
                    #self.printer.print_var("relax_list",relax_list)
                    #self.printer.print_var("relax_vals",relax_vals)
                #if ((relax_fc < 1e-12) and (relax_fd < 1e-12)):
                    #break
                if (relax_fc < relax_fd):
                    b = d
                    d = c
                    relax_fd = relax_fc
                    need_update_c = True
                    need_update_d = False
                elif (relax_fc >= relax_fd):
                    a = c
                    c = d
                    relax_fc = relax_fd
                    need_update_c = False
                    need_update_d = True
                else: assert(0)
                if (relax_k >= self.relax_n_iter_max):
                #if (relax_k >= 9) and (numpy.argmin(relax_vals) > 0):
                    break
                relax_k += 1
            self.printer.dec()
            self.problem.sol_func.vector().axpy(
                -cur,
                self.problem.dsol_func.vector())
            if (len(self.problem.subsols) > 1):
                dolfin.assign(
                    self.problem.get_subsols_func_lst(),
                    self.problem.sol_func)
            #self.printer.print_var("relax_vals",relax_vals)

            self.relax = relax_list[numpy.argmin(relax_vals)]
            self.printer.print_sci("relax",self.relax)
            if (self.relax == 0.):
                self.printer.print_str("Warning! Optimal relaxation is null…")

    def compute_relax_backtracking(self):
        k_relax = 1
        self.printer.inc()
        while (True):
            relax = 1./self.relax_backtracking_factor**(k_relax-1)
            self.problem.sol_func.vector().axpy(relax, self.problem.dsol_func.vector())
            self.assemble_linear_system()
            res_is_finite = numpy.isfinite(self.res_vec).all()
            # print("numpy.isfinite(self.res_vec).all()", res_is_finite)
            self.problem.sol_func.vector().axpy(-relax, self.problem.dsol_func.vector())
            if (res_is_finite):
                self.relax = relax
                break
            if (k_relax == self.relax_n_iter_max):
                self.relax = 0.
                self.printer.print_str("Warning! Optimal relaxation is null…")
                break
            k_relax += 1
        self.printer.dec()


    def update_sol(self):

        # for constraint in self.problem.constraints+self.problem.steps[k_step-1].constraints:
        #     print(constraint.bc.get_boundary_values())
        self.problem.sol_func.vector().axpy(
            self.relax,
            self.problem.dsol_func.vector())
        # for constraint in self.problem.constraints+self.problem.steps[k_step-1].constraints:
        #     print(constraint.bc.get_boundary_values())
        # self.printer.print_var("sol_func",self.problem.sol_func.vector().get_local())

        if (len(self.problem.subsols) > 1):
            dolfin.assign(
                self.problem.get_subsols_func_lst(),
                self.problem.sol_func)
            # for subsol_name,subsol in self.problem.subsols.items()):
            #     self.printer.print_var(subsol_name+"_func",subsol.func.vector().get_local())



    def compute_sol_norm(self):

        self.subsol_norm_lst = [subsol.func.vector().norm("l2") for subsol in self.problem.subsols.values()]
        self.subsol_norm_old_lst = [subsol.func_old.vector().norm("l2") for subsol in self.problem.subsols.values()]
        for (k_subsol,subsol) in enumerate(self.problem.subsols.values()):
            self.printer.print_sci(subsol.name+"_norm"    ,self.subsol_norm_lst[k_subsol]    )
            self.printer.print_sci(subsol.name+"_norm_old",self.subsol_norm_old_lst[k_subsol])



    def compute_sol_err(self):

        self.subsol_err_lst = [dmech.compute_error(
            val=self.dsubsol_norm_lst[k_subsol],
            ref=max(
                self.subsol_norm_lst[k_subsol],
                self.subsol_norm_old_lst[k_subsol])) for k_subsol in range(len(self.problem.subsols))]
        for (k_subsol,subsol) in enumerate(self.problem.subsols.values()):
            self.printer.print_sci(subsol.name+"_err",self.subsol_err_lst[k_subsol])



    def exit_test(self):

        self.success = all([self.subsol_err_lst[k_subsol]<self.sol_tol[k_subsol] for k_subsol in range(len(self.problem.subsols)) if self.sol_tol[k_subsol] is not None])



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin

import dolfin_mech as dmech
from .Operator import Operator

################################################################################

class HyperElasticityOperator(Operator):

    def __init__(self,
            U,
            U_test,
            kinematics,
            material_model,
            material_parameters,
            measure,
            formulation="PK1"): # PK1 or PK2 or ener

        self.kinematics = kinematics
        self.material   = dmech.material_factory(kinematics, material_model, material_parameters)
        self.measure    = measure

        assert (formulation in ("PK1", "PK2", "ener")),\
            "\"formulation\" should be \"PK1\", \"PK2\" or \"ener\". Aborting."

        if (formulation == "ener"):
            self.res_form = dolfin.derivative(self.material.Psi, U, U_test) * self.measure
        elif (formulation == "PK2"):
            dE_test = dolfin.derivative(
                self.kinematics.E, U, U_test)
            self.res_form = dolfin.inner(self.material.Sigma, dE_test) * self.measure
        elif (formulation == "PK1"):
            dF_test = dolfin.derivative(
                self.kinematics.F, U, U_test)
            self.res_form = dolfin.inner(self.material.P, dF_test) * self.measure



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
###                                                                          ###
### And Cécile Patte, 2019-2021                                              ###
###                                                                          ###
### INRIA, Palaiseau, France                                                 ###
###                                                                          ###
################################################################################

import dolfin

import dolfin_mech as dmech
from .Material_Elastic import ElasticMaterial

################################################################################

class PorousElasticMaterial(ElasticMaterial):



    def __init__(self,
            solid_material,
            scaling="no",
            Phis0=None):

        self.solid_material = solid_material

        if (scaling == "no"):
            scaling = dolfin.Constant(1)
            # self.Psi   = self.material.Psi
            # if (hasattr(self.material,       "Sigma")): self.Sigma       = self.material.Sigma
            # if (hasattr(self.material,           "P")): self.P           = self.material.P
            # if (hasattr(self.material,       "sigma")): self.sigma       = self.material.sigma
            # if (hasattr(self.material, "dWbulkdPhis")): self.dWbulkdPhis = self.material.dWbulkdPhis
        elif (scaling == "linear"):
            scaling = Phis0
            # self.Psi   = Phis0 * self.material.Psi
            # if (hasattr(self.material,       "Sigma")): self.Sigma       = Phis0 * self.material.Sigma
            # if (hasattr(self.material,           "P")): self.P           = Phis0 * self.material.P
            # if (hasattr(self.material,       "sigma")): self.sigma       = Phis0 * self.material.sigma
            # if (hasattr(self.material, "dWbulkdPhis")): self.dWbulkdPhis = Phis0 * self.material.dWbulkdPhis
        else:
            assert (0),\
                "scaling must be \"no\" or \"linear\". Aborting."

        for attr in ("Psi", "Sigma", "P", "sigma", "dWbulkdPhis", "dWporedPhif"):
            if (hasattr(self.solid_material, attr)):
                setattr(self, attr, scaling * getattr(self.solid_material, attr))



#coding=utf8

################################################################################
###                                                                          ###
### Created by Alice Peyraut, 2023-2024                                      ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
###                                                                          ###
### And Martin Genet, 2018-2024                                              ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin

import dolfin_mech as dmech
from .Operator import Operator

################################################################################

class PressureBalancingGravityLoadingOperator(Operator):

    def __init__(self,
            X,
            x0,
            x0_test,
            lbda,
            lbda_test,
            mu,
            mu_test,
            p,
            p_test,
            gamma,
            gamma_test,
            U,
            U_test,
            Phis0,
            rho_solid,
            kinematics,
            N,
            dS,
            dV,
            breathing_constant,
            P0_val=None, P0_ini=None, P0_fin=None,
            f_val=None, f_ini=None, f_fin=None):

        self.measure = dV
        self.dS = dS

        self.V0 = dolfin.assemble(dolfin.Constant(1)*self.measure)
        self.rho0 = dolfin.assemble(Phis0 * self.measure)

        self.tv_f = dmech.TimeVaryingConstant(
            val=f_val, val_ini=f_ini, val_fin=f_fin)
        f = self.tv_f.val

        self.tv_P0 = dmech.TimeVaryingConstant(
            val=P0_val, val_ini=P0_ini, val_fin=P0_fin)
        P0 = self.tv_P0.val

        self.tv_grad = dmech.TimeVaryingConstant(
            val=None, val_ini=f_ini[2], val_fin=f_fin[2]) #### AP2023 - depends on the orientation of gravity: here, g = [0., 0., 9.81e3]
        grad = self.tv_grad.val

        nf = dolfin.dot(N, dolfin.inv(kinematics.F))
        nf_norm = dolfin.sqrt(dolfin.inner(nf,nf))
        n = nf/nf_norm

        x = X + U
        x_tilde = x-x0

        P_tilde = P0 + dolfin.Constant(rho_solid) * grad * ( x[2]- x0[2]) - dolfin.Constant(breathing_constant)*dolfin.Constant(rho_solid) * grad * ( x[1]- x0[1]) * ( x[1]- x0[1]) 

        grads_p = dolfin.dot(dolfin.grad(p-P_tilde), dolfin.inv(kinematics.F)) - n*(dolfin.dot(n,dolfin.dot(dolfin.grad(p-P_tilde), dolfin.inv(kinematics.F))))
        grads_p_test = dolfin.dot(dolfin.grad(p_test), dolfin.inv(kinematics.F)) - n*(dolfin.dot(n,dolfin.dot(dolfin.grad(p_test), dolfin.inv(kinematics.F))))

        self.res_form = dolfin.Constant(1e-8)*p*p_test * kinematics.J * self.measure
        self.res_form -= dolfin.inner(rho_solid*Phis0*f, U_test) * self.measure
        self.res_form -=  dolfin.inner(-p * n, U_test) * nf_norm * kinematics.J * dS
        self.res_form += dolfin.inner(rho_solid*Phis0*f, lbda_test) * self.measure
        self.res_form += dolfin.inner(-p * n, lbda_test) * nf_norm * kinematics.J * dS
        self.res_form += - dolfin.dot(lbda, n) * p_test * nf_norm * kinematics.J * self.dS
        self.res_form += - dolfin.dot(mu, dolfin.cross(x_tilde, n)) *  p_test * nf_norm * kinematics.J * self.dS
        self.res_form += gamma  *  p_test * nf_norm * kinematics.J * self.dS
        self.res_form +=  dolfin.inner(grads_p, grads_p_test) * nf_norm * kinematics.J * self.dS
        self.res_form += dolfin.inner(dolfin.cross(x_tilde, -p * n), mu_test) * nf_norm * kinematics.J * dS
        self.res_form += (p-P_tilde)*gamma_test * nf_norm * kinematics.J * dS
        self.res_form -= dolfin.inner((Phis0 * x / dolfin.Constant(self.rho0) - x0/self.V0), x0_test) * self.measure


    def set_value_at_t_step(self,
            t_step):

        self.tv_f.set_value_at_t_step(t_step)
        self.tv_P0.set_value_at_t_step(t_step)
        self.tv_grad.set_value_at_t_step(t_step)

################################################################################

class PressureBalancingGravity0LoadingOperator(Operator):

    def __init__(self,
            x,
            x0,
            u_test,
            lbda,
            lbda_test,
            mu,
            mu_test,
            p,
            p_test,
            gamma,
            gamma_test,
            rho_solid,
            phis,
            n,
            dS,
            dV,
            breathing_constant,
            P0_val=None, P0_ini=None, P0_fin=None,
            f_val=None, f_ini=None, f_fin=None):

        self.dS = dS
        self.measure = dV

        self.tv_f = dmech.TimeVaryingConstant(
            val=f_val, val_ini=f_ini, val_fin=f_fin)
        f = self.tv_f.val

        self.tv_P0 = dmech.TimeVaryingConstant(
            val=P0_val, val_ini=P0_ini, val_fin=P0_fin)
        P0 = self.tv_P0.val

        self.tv_grad = dmech.TimeVaryingConstant(
            val=None, val_ini=0, val_fin=f_fin[2]) #### AP2023 - depends on the orientation of gravity: here, g = [0., 0., 9.81e3]
        grad = self.tv_grad.val

        x_tilde = x-dolfin.Constant(x0)

        P_tilde = P0 + dolfin.Constant(rho_solid) * grad * ( x[2]- dolfin.Constant(x0[2])) - dolfin.Constant(breathing_constant)*dolfin.Constant(rho_solid) * grad* ( x[1]- dolfin.Constant(x0[1])) * ( x[1]- dolfin.Constant(x0[1])) 
        
        grads_p = dolfin.grad(p-P_tilde) - n*(dolfin.dot(n,dolfin.grad(p-P_tilde)))
        grads_p_test = dolfin.grad(p_test) - n*(dolfin.dot(n,dolfin.grad(p_test)))

        self.res_form = dolfin.Constant(1e-8)*p*p_test * self.measure
        self.res_form -= dolfin.inner(rho_solid*phis*f, u_test) * self.measure
        self.res_form -=  dolfin.inner(-p * n, u_test) * dS
        self.res_form += dolfin.inner(rho_solid*phis*f, lbda_test) * self.measure
        self.res_form += dolfin.inner(-p * n, lbda_test) * dS
        self.res_form += - dolfin.dot(lbda, n) *  p_test * self.dS
        self.res_form += - dolfin.dot(mu, dolfin.cross(x_tilde, n)) *  p_test * self.dS
        self.res_form += gamma  *  p_test * self.dS
        self.res_form +=  dolfin.inner(grads_p, grads_p_test) * self.dS
        self.res_form += dolfin.inner(dolfin.cross(x_tilde, -p * n), mu_test) * dS
        self.res_form += (p-P_tilde)*gamma_test * dS

    def set_value_at_t_step(self,
            t_step):
        self.tv_f.set_value_at_t_step(t_step)
        self.tv_P0.set_value_at_t_step(t_step)
        self.tv_grad.set_value_at_t_step(t_step)



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin
import gmsh
import meshio

################################################################################

def run_Ball_Mesh(
        params={}):

    X0 = params.get("X0", 0.5)
    Y0 = params.get("Y0", 0.5)
    Z0 = params.get("Z0", 0.5)
    R  = params.get("R" , 0.3)
    l  = params.get("l" , 0.01)

    mesh_filebasename = params.get("mesh_filebasename", "mesh")

    ################################################################### Mesh ###
    
    gmsh.initialize()
    gmsh.clear()
    factory = gmsh.model.occ

    sp = factory.addSphere(xc=X0, yc=Y0, zc=Z0, radius=R)

    factory.synchronize()

    ps = gmsh.model.addPhysicalGroup(dim=3, tags=[sp])

    mesh_gmsh = gmsh.model.mesh

    mesh_gmsh.setSize(gmsh.model.getEntities(0), l)
    mesh_gmsh.generate(dim=3)

    gmsh.write(mesh_filebasename+".vtk")

    gmsh.finalize()

    mesh_meshio = meshio.read(mesh_filebasename+".vtk")

    meshio.write(mesh_filebasename+".xdmf", mesh_meshio)

    mesh = dolfin.Mesh()
    dolfin.XDMFFile(mesh_filebasename+".xdmf").read(mesh)

    dolfin.File(mesh_filebasename+".xml") << mesh

    ############################################################# Boundaries ###

    boundaries_mf = dolfin.MeshFunction(
        value_type="size_t",
        mesh=mesh,
        dim=1)

    boundaries_mf.set_all(0)

    S_sd = dolfin.AutoSubDomain(
        lambda x, on_boundary:
            on_boundary and\
            dolfin.near(
                (x[0]-X0)**2 + (x[1]-Y0)**2 + (x[2]-Z0)**2,
                R**2,
                eps=1e-3))

    S_id = 1; S_sd.mark(boundaries_mf, S_id)

    # dolfin.XDMFFile(mesh_filebasename+"-boundaries.xdmf").write(boundaries_mf)

    ################################################################# Return ###

    return mesh, boundaries_mf, S_id



#coding=utf8

################################################################################
###                                                                          ###
### Created by Martin Genet, 2018-2024                                       ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
###                                                                          ###
### And Cécile Patte, 2019-2021                                              ###
###                                                                          ###
### INRIA, Palaiseau, France                                                 ###
###                                                                          ###
###                                                                          ###
### And Colin Laville, 2021-2022                                             ###
###                                                                          ###
### École Polytechnique, Palaiseau, France                                   ###
###                                                                          ###
################################################################################

import dolfin

import dolfin_mech as dmech
from .Material_Elastic import ElasticMaterial

################################################################################

class WporeLungElasticMaterial(ElasticMaterial):



    def __init__(self,
            Phif,
            Phif0,
            parameters):

        assert ('eta' in parameters)
        self.eta = dolfin.Constant(parameters['eta'])

        self.n = dolfin.Constant(parameters.get('n', 1))
        self.p = dolfin.Constant(parameters.get('p', 1))
        self.q = dolfin.Constant(parameters.get('q', 1))

        Phif = dolfin.variable(Phif)
        r = Phif/Phif0
        r_inf = Phif0**(self.p-1)
        r_sup = Phif0**(1/self.q-1)
        self.Psi = self.eta * dolfin.conditional(dolfin.lt(r, 0.), r/dolfin.Constant(0.), dolfin.conditional(dolfin.lt(r, r_inf), (r_inf/r - 1)**(self.n+1), dolfin.conditional(dolfin.gt(r, r_sup), (r/r_sup - 1)**(self.n+1), 0)))
        self.dWporedPhif = dolfin.diff(self.Psi, Phif)
